/* microabc
 * Generates a file for the preprocessor abcpp, with MIDI mapping or
 * ABC microtonal notation, intended to notate arbitrary scales and
 * tunings.
 */

/* microabc
 * Gera um arquivo para o pr-processador abcpp, com mapeamento MIDI
 * ou notao microtonal ABC, para notao de escalas e afinaes
 * arbitrrias.
 */

/*
 * Copyright (C) 2006-2008 Hudson Flvio Meneses Lacerda <hfml @ brfree.com.br>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* SUJEITO  LICENA PBLICA GERAL DO GNU (GNU GPL)
 * TEXTO DA LICENA EM:
 *	ARQUIVO "Licenca.txt"
 *      <http://www.gnu.org/copyleft/gpl.html>
 *      <http://www.magnux.org/doc/GPL-pt_BR.txt>
 */


/***************************************************
 * TODO:
 * - A LOT OF TESTINGS!!!
 * - Output formats: Pitch numbers and Octave.PitchClass
 *
 ****************************************************/

#include "microabc.h"
#include "interval.h"
#include "sagittal.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h> /* link to it with the gcc option -lm */

//extern const int primes[NPRIMES]; /*  */
extern int sagoctave[2]; /* sagittal.c */

char aliases[MAX_MACROS][64];	/* List of aliases (macro names) */

char macrodef[MAX_DEFINITIONS][2][64];	/* List of actual macro
					   definitions for internal
					   basic preprocessor */
int nmacrodefs; /* number of used definitions */

double temper[NPRIMES];  	/* prime factors temperings in cents */

char comment[2][64];		/* comment markers */

int outputmode;                 /* output mode: literal, chromatic,
				   diatonic or microtonal */
#define LITERAL 1
#define CHROMATIC  2
#define DIATONIC 3
#define MICROTONAL 4

int basepitch;			/* base pitch for microtonal scale */
double basefreq;		/* base frequency for microtonal scale */
double sclmodulo;		/* modulo (formal octave) in cents */
double a_freq=440.0;		/* "A" frequency */
double scale[MAX_SCALE_SIZE];	/* scale data */
int equalt;			/* Compute equaltemp (bool) */
int lineart;			/* Compute lineartemp -------------
				   When lineart>0, this variable
				   stores the 'count downwards' value.
				   Else If lineart==0, don't compute scale.
				   Else lineart<0, then set count downwards=0.
				*/
double generator;		/* Generator for linear temperament
				   (formal fifth) */

int nnumerators;		/* Number of numerators */
int numerators[MAX_NUMERATORS];	/* Numerators for microtonal accidentals */
int denominator;		/* Denominador for microtonal accidental */
int abcm2ps_den;		/* Denominator for abcm2ps quantization */
int abcm2ps_reduction;		/* Reduction of fractions to 1/2 (bool) */
int abcm2ps_psacc;	/* Output PS code for eighth-tone accidentals */
int scl;		/* index of scala-like input column (interval) */
int abcalias[MAX_ALIAS];	/* indexes for abc macro names */
int sagittalias[MAX_ALIAS];	/* indexes for Sagittal macro names */
int keymodulation;                    /* rotate the function of a scale
				 degree, i.e., selects a mode of the
				 scale.  Similar to the command `key'
				 of the program Scala, */

double freq=0.0;
int outputfreq;			/* Output frequencies in microtonal mode */
int accid_dmacros;		/* Output dumb macros for accidentals */
char accid_def[MAX_ACCID][2][64];  /* Accidental definitions */
int naccid;			/* Number of accidental definitions */
int abcnoaccid;			/* Don't output accidentals in replaceabc */
int sagittalnoaccid;		/* Don't output accidentals in replacesagittal */
int sagittalvisibleaccid;	/* Output visible accidentals in replacesagittal */
int usingsagittalalias;		/* bool */
int avoidmixedsag;              /* workaround to avoid mixed-flat in pure Sagittal */
double sagittal_basefreq;          /* reference frequency for Sagittal */
int inspect;                    /* print details (e.g. Sagittal to ABC converting) */
char sagittal_base[2];          /* nominal to tune Sagittal */

/* maybe better a struct? */
int sagittal2scale;
double sag_generator;
int sag_equalt;
int sag_lineart;
int sag_modulo;
double sag_sclmodulo;
int embed_sagittal_fmt;

int centro;			/* MIDI number for pitch 0 (center pitch) */
int modulo;			/* Number of degrees for modulo instance */
int force_modulo;		/* Don't count scale items */
int ambito[2];		/* Range relative to center pitch */
int amin, amax;			/* Lower and higher MIDI pitches */
int transp;			/* Transposition interval */

int indalias[MAX_ALIAS];	/* Index of columns used for aliases */
char up[8];			/* Additional symbol for high pitches */
char down[8];			/* Additional symbol for low pitches */
int fix_updown;			/* Try remove redundant up/down symbols */
char leftd[8];			/* Left delimiter for macro names */
char rightd[8];			/* Right delimiter for macro names */
int oitavanum;		/* Represent modulo instance as numbers (bool) */
int oitavasep;		/* Put a separator before modulo instance number (bool) */
char oitavasepchar;  /* Separator for modulo instance number */
int undef;			/* Generate output to undefine macros (bool) */
int chroma;			/* Chromatic ABC output (bool) */
int replace;		/* Index of input column for replacement */
			/* negative value for no replacement column */
int replace_updown;	/* Put octave indication in replacement mode (bool) */
int replace_abc;	/* abc replacement (bool) */
int replace_sagittal;	/* sagittal replacement (bool) */
int replace_nominal;	/* abc|sagittal nominal replacement (number of nominals) */
int nominal_offset;     /* reference (diatonic) pitch for nomimal replacement, as number */
char skipstring[64];		/* skip column string */
char textbuf[MAX_BUFLINES][MAX_BUFLINELEN];	/* buffer for echo: command */
int textbuflen = 0;		/* number of textbuf lines */

char forpsbuf[MAX_BUFLINES][MAX_BUFLINELEN]; /* buffer for FOR_PS: command */
int forpsbuflen = 0;		/* number of lines */

char formidibuf[MAX_BUFLINES][MAX_BUFLINELEN];	/* buffer for FOR_MIDI: command */
int formidibuflen = 0;		/* number of lines */

int ignore;
int processfor;			/* FOR_MIDI or FOR_PS */
int procsag;			/* Whether process sagittal pitches or not */
int procmac;			/* Whether process defined macros or not */

int microtonal;			/* microtonal mode (bool) */

int timidity;			/* Output timidity++ frequencies (bool) */
int outscala;			/* Output scala *.scl file (bool) */
int outscalakbm;		/* Output scala *.kbm file (bool) */
int snt;   /* Output snt file for abc2midi (1.80 January 29 2006) */
int inputscala;			/* input scala file (bool) */

char sname[MAX_NAME_LENGHT+1];	       /* To output #define <name> "" */
char snt_file[MAX_NAME_LENGHT+1];	/* File name for snt data */
char scala_input_file[MAX_NAME_LENGHT+1]; /* input scala file name */
char infile[MAX_NAME_LENGHT+1];
char outfile[MAX_NAME_LENGHT+1];
char abpfile[MAX_NAME_LENGHT+1];
char testfilename[MAX_NAME_LENGHT+1];
char timidity_file[MAX_NAME_LENGHT+1];	/* File name for timidity data */
char scalakbm_file[MAX_NAME_LENGHT+1];	/* Scala *.kbm file name */
char scala_file[MAX_NAME_LENGHT+1];	/* Scala *.scl file name */

int easysag;

char commandline[BUFSIZ];
int inputfromfile=0, outputtofile=0, inputabpfile=0;
int generate_testfile=0;

int main(int argc, char **argv)
{
    char s[BUFSIZ];
    int i;
    FILE *fp=NULL, *fp2=NULL;

    initialize(); /* default values */

    commandline[0]='\0';
    for(i=0;i<argc;i++) {
	strncat(commandline,argv[i],BUFSIZ-1-strlen(commandline));
	if(i<argc-1)
	    strncat(commandline," ",BUFSIZ-1-strlen(commandline));
    }

    /* read a few command line options */
    for(i=1;i<argc;i++) {
	sprintf(s,"%s",argv[i]);
	if(s[0]=='-') {
	    if(s[1]=='E') { /* ``Easy Sagittal'' flag: */
		/* It is the same as: -i- -S -e -oEasyFile.abc */
		easysag = 1;
		inputfromfile=1;
		strcpy(infile,"-");
		embed_sagittal_fmt=1;
		procsag=1; 
		procmac=0;
		if(outputtofile==0) {
		    strcpy(outfile,"EasyFile.abc");
		    outputtofile=1;
		}
	    }
	    else if(s[1]=='i') {
		strncpy(infile,s+2,MAX_NAME_LENGHT);
		inputfromfile=1;
	    }
	    else if(s[1]=='I') {
		inspect=1; /* inspect */
	    }
	    else if(s[1]=='P') { /* NOTE: this code just detects the
				    presence of -P */
		processfor=FOR_PS;
	    }
	    else if(s[1]=='M' || s[1]=='p') { /* NOTE: this code just
						 detects the presence
						 of -M or -p */
		processfor=FOR_MIDI;
	    }
	    else if(s[1]=='h') { /* print info and quit */
		printf("\nmicroabc (C) 2006-2008 Hudson Lacerda"
		       " <hfml@brfree.com.br>\n"
		       "\nThis program is free software "
		       "released under the GNU General Public License.\n"
		       "There is no warranty. "
		       "Licensing terms available in the file \"License\".\n"
		       "\nBasic usage:\n"
		       "    microabc < infile > outfile\n"
		       "\nFor more information on usage, "
		       "please read the file \"microabc.txt\".\n\n"
		       );
		exit(0);
	    }
	}
    }

    /* input from infile or stdin */
    ///// TODO: function input() should do all this:
    if(inputfromfile) {
 	if(strcmp(infile,"-")==0) { /* No input with -i- from command line */
	    if(force_modulo==0) {
		if(inputabpfile)
		    set_pythagorean(); /* pythagorean for Sagittal */
		else
		    set_meantone(); /// TODO: choose what to do with this!
	    }
	}
	else if((fp=fopen(infile,"rt"))!=NULL) { /* Input from infile */
	    if(!input(fp)) {
		fprintf(stderr,"Warning: no input data in %s!!!\n",
			infile);
		fclose(fp);
		exit(1);
	    }
	    fclose(fp);
	}
	else {
	    fprintf(stderr,"ERROR: Cannot open input (-i) file \"%s\"\n\n",
		    infile);
	    exit(1);
	}
    }
    else if(!input(stdin)) {
	fprintf(stderr,"Warning: no input data!!!\n");
	if(force_modulo==0) {
	    if(inputabpfile)
		set_pythagorean(); /* pythagorean for Sagittal */
	    else
		set_meantone();
	}
    }

    /* Get (remaining) options from command line  */
    if(argc>1) {
	for(i=1;i<argc;i++) {
	    sprintf(s,"%s",argv[i]);
	    if(s[0]=='-') {
		if(s[1]=='E') { /* ``Easy Sagittal'' flag */
		    /* Was set before */
		}
		else if(s[1]=='e') {
		    embed_sagittal_fmt=1;
		}
		else if(s[1]=='i') {
		    /* input file already read. */
		}
		else if(s[1]=='I') {
		    /* Was set before. */
		}
		else if(s[1]=='o') {
		    if(s[2]=='-') {
			outputtofile=0;
			outfile[0]='\0';
		    }
		    else {
			strncpy(outfile,s+2,MAX_NAME_LENGHT);
			outputtofile=1;
		    }
		}
		else if(s[1]=='t') {
		    strcpy(testfilename,s+2);
		    if(strlen(testfilename)<1 ||espacos(testfilename))
			strcpy(testfilename,"testfile.abp");
		    generate_testfile=1;
		}
		else if(s[1]=='p' || s[1]=='M') {
		    /* use basic preprocessor for abc2MIDI */
		    strncpy(abpfile,s+2,MAX_NAME_LENGHT);
		    inputabpfile=1;
		    processfor=FOR_MIDI;
		}
		else if(s[1]=='P') {
		    /* use basic preprocessor for abcm2PS */
		    strncpy(abpfile,s+2,MAX_NAME_LENGHT);
		    inputabpfile=1;
		    processfor=FOR_PS;
		}
		else if(s[1]=='S') {/* control Sagittal preprocessor */
		    if(s[2]=='0') {
			procsag=0; 
			procmac=1;
		    }
		    else if(s[2]=='1') {
			procsag=1; 
			procmac=0;
		    }
		    else if(s[2]=='2') {
			procsag=procmac=1;
		    }
		    else {
			procsag=1;
			procmac=0;
		    }
		}
		else {
		    fprintf(stderr,"\nERROR: invalid option %s\n\n",s);
		    exit(-1);
		}
	    }
	    else {
		if(!get_option(s)) {
		    fprintf(stderr,"\nERROR: invalid option %s\n\n",s);
		    exit(-1);
		}
	    }
	}
    }

    fprintf(stderr,"#\n# microabc (C) 2006-2008 Hudson Lacerda\n#\n");

    prepare(); /// TODO: remove prepare() from output().

    /* generate output to fileout */
    if(outputtofile) {
	if((fp=fopen(outfile,"wt"))!=NULL)
	    {
		output(fp);
		fclose(fp);
		if(inputabpfile) {
		    /* Generate test file before preprocessing (it) */
		    if(generate_testfile)
			output_testfile(testfilename);
		    /* Preprocess abp file  */
		    if((fp2=fopen(abpfile,"rt"))!=NULL) {
			fp=fopen(outfile,"awt"); /* append */
			preprocess(fp2,fp,processfor);
			fclose(fp);
			fclose(fp2);
		    }
		    else { /* TODO: centralize this code */
			fprintf(stderr,
				"\n\nERROR: Cannot open abp (-p/-M) file \"%s\"\n\n",
				abpfile);	    
			exit(-1);
		    }
		}
		fprintf(stderr,"OUTPUT FILE GENERATED: `%s'\n",outfile);
	    }
	else {
	    fprintf(stderr,
		    "\n\nERROR: Cannot open output (-o) file \"%s\"\n\n",
		    outfile);
	    exit(-1);
	}
    }
    /* generate output to stdout */
    else {
	output(stdout);
	if(inputabpfile) {
	    /* Generate test file before preprocessing (it) */
	    if(generate_testfile)
		output_testfile(testfilename);
	    /* Preprocess abc file */
	    if((fp2=fopen(abpfile,"rt"))!=NULL) {
		preprocess(fp2,stdout, processfor);
		fclose(fp2);
	    }
	    else { /* TODO: centralize this code */
		fprintf(stderr,
			"\n\nERROR: Cannot open abp (-p/M) file \"%s\"\n\n",
			abpfile);	    
		exit(-1);
	    }
	}
    }

    /* Generate test file when not preprocessing */
    if(inputabpfile==0 && generate_testfile==1)
	output_testfile(testfilename);
   
    return 0;
}

int initialize() /* set default/initial values for variables */
{
    int i;

    replace_nominal=0;
    nominal_offset=0;
    easysag=0;
    inputfromfile=0;
    outputtofile=0;
    inputabpfile=0;
    generate_testfile=0;
    infile[0]=outfile[0]='\0';

    outputmode=CHROMATIC;
    ignore=0;
    embed_sagittal_fmt=0;
    generate_testfile=0;
    avoidmixedsag=1;
    strncpy(sagittal_base,"D",1);
    sagittal_basefreq=midi2freq(62,440.0);
    inspect=0;
    sagoctave[0] = SAG_DEFAULT_OCT;
    sagoctave[1] = 0;
    usingsagittalalias=0;
    sname[0]='\0';
    comment[0][0]='{';comment[0][1]='\0';
    comment[1][0]='}';comment[1][1]='\0';
    centro=60;
    modulo=0;
    force_modulo=0;
    scl= -1;
    microtonal=0;
    sclmodulo= 1200.0;
    equalt= 0;
    lineart = 0;
    generator = interv2cents(3.0/2.0);
    processfor=FOR_MIDI;
    procsag=0;
    procmac=1;
    freq=0.0;
    a_freq= 440.0;
    basepitch = 0;
    basefreq=midi2freq(60,440.0);
    outputfreq=0;
    nnumerators = 0;
    denominator= 4096;
    abcm2ps_den=0;
    abcm2ps_reduction=2;
    abcm2ps_psacc=1;
    timidity= 0;
    outscala= 0;
    outscalakbm= 0;
    snt= 0;
    inputscala= 0;
    accid_dmacros= 0;
    naccid= 0;
    abcnoaccid=0;
    sagittalnoaccid=0;
    sagittalvisibleaccid=0;
    ambito[0]= -60; ambito[1]= 67;
    amin= 48; amax= 72; /* automatically set from ambito[] by input() */
    transp= 0;
    keymodulation= 0;
    up[0]='\''; up[1]='\0';
    down[0]=','; down[1]='\0';
    fix_updown=1;
    leftd[0]='['; leftd[1]='\0';
    rightd[0]=']'; rightd[1]='\0';
    skipstring[0]='.'; skipstring[1]='\0';
    oitavanum= 0;
    oitavasep= 1;
    oitavasepchar=OCTSEP;
    undef= 0;
    chroma= 1;
    replace= -1;
    replace_updown=0;
    replace_abc=0;
    sagittal2scale = 0;
    sag_equalt = equalt;
    sag_lineart = lineart;
    sag_modulo = modulo;
    sag_sclmodulo = sclmodulo;
    sag_generator=generator;

    for(i=0;i<MAX_SCALE_SIZE;i++)
	scale[i]=0.0;

    for(i=0;i<MAX_BUFLINES;i++) {
	textbuf[i][0]='\0';
	forpsbuf[i][0]='\0';
	formidibuf[i][0]='\0';
    }
    textbuflen=0;
    forpsbuflen=0;
    formidibuflen=0;

    for(i=0;i<MAX_ALIAS;i++)
	indalias[i]=-1;

    for(i=0;i<MAX_ALIAS;i++)
	abcalias[i]=-1;

    for(i=0;i<MAX_ALIAS;i++)
	sagittalias[i]=-1;

    for(i=0;i<NPRIMES;i++)
	temper[i]=0.0;

    for(i=0;i<MAX_DEFINITIONS;i++) {
	macrodef[i][0][0]='\0'; /* name */
	macrodef[i][1][0]='\0'; /* replacement */
    }
    nmacrodefs=0;

    return 1;
}

int input(FILE *fp)
{
    char s[BUFSIZ];
    int i,c;
    char a[12], b[12];
    int tmpmodulo;
 
    i=0;
    while(fgetline(s,128,fp) != NULL
	  && i<MAX_MACROS) {
	if(!get_option(s)) /* if unknown command */
	    /* then is scale degree */
	    if(strlen(s)>0 && !espacos(s)) { /* load aliases columns */
		strncpy(aliases[i],s,sizeof(aliases[0]));
		i++;
	    }
    }
    c=i;



    /* auto adjusts (maybe put warnings/errors would be better ??) */

    /* generate scale */

    if(scl>=0) {
	tmpmodulo=modulo;
	modulo=MAX_SCALE_SIZE;
	strcpy(a,up); /* backup updown */
	strcpy(b,down);
	up[0]=down[0]='\0';
	for (i=0; i<MAX_SCALE_SIZE; i++) {
	    if (i>c)
		break;	    
	    alias(s,i,scl);
	    scale[i]=scl2cents(s);
	}
	strcpy(up,a); /* restore updown */
	strcpy(down,b);
	if(c>=MAX_SCALE_SIZE) {
	    fprintf(stderr,"\nERROR: SCALE OVERFULL\n\n");
	}
	modulo=tmpmodulo;
    }

    i=c;

    if(force_modulo==0)
	modulo=i;
    else if(modulo>i || modulo<1) {
	modulo=i;
	fprintf(stderr,
		"\nERROR: BAD `size:' (`mod:') RESET TO %d\n\n",i);
    }
    modulo= modulo<MAX_SCALE_SIZE?modulo:MAX_SCALE_SIZE;

    return i;
}


int get_option(char *s) /* TODO: fixme!!! WARNING: s[] is changed!!! */
{
    char t[BUFSIZ],u[BUFSIZ],*p;
    int c, x, d;
    static int whichbuf = 0;
#define PS_BUF 0
#define MIDI_BUF 1

    static char abcnominals[]="CDEFGABcdefgab";
    char *pointer;

    /* discard line with spaces only */
    if(espacos(s))
	return 0;

    /* remove blank spaces */

    strcpy(t,s);

    p=t;
    while(isspace(*p) && *p!='\0')
	p++;
    strncpy(s,p,BUFSIZ-1);

    /* discard EOL characters */
    //    p=strchr(s,'\n'); if(p!=0) *p = '\0'; /* descarta '\n' (DOS) */
    
    p=strchr(s,13); if(p!=0) *p = '\0'; /* descarta eol */
    p=strchr(s,10); if(p!=0) *p = '\0'; /* descarta eol */    

    /* Try to interpret string */

    //    fprintf ( stderr, "s=`%s'\n", s );

    /* different sections for .mid or for .ps */

    if(sscanf(s,"outputfor: %s",t)>0) {
	if(strcmp("PS",t)==0)
	    processfor=FOR_PS;
	else if(strcmp("MIDI",t)==0)
	    processfor=FOR_MIDI;
	else
	    fprintf(stderr,"\nERROR: Invalid option in `outputfor:' %s\n\n",t);
	return 1;
    }
    if(sscanf(s,"FOR_MIDI: %s",t)>0) {
	if(strcmp("BEGIN",t)==0) {
	    whichbuf = MIDI_BUF; /* points to options for midi */
	    /// if(processfor!=FOR_MIDI)
	    ignore=1; /* store */
	}
	else if(strcmp("END",t)==0) {
	    ignore=0;
	}
	else {
	    fprintf(stderr,"\nERROR: Invalid option in FOR_MIDI:"
		    " `%s' (Must be 'BEGIN' or 'END'\n\n",t);
	    exit(-1);
	}
	return 1;
    }
    if(sscanf(s,"FOR_PS: %s",t)>0) {
	if(strcmp("BEGIN",t)==0) {
	    whichbuf = PS_BUF; /* points to options for ps  */
	    //// if(processfor!=FOR_PS)
	    ignore=1; /// in reality, store
	}
	else if(strcmp("END",t)==0) {
	    ignore = 0;
	}
	else {
	    fprintf(stderr,"\nERROR: Invalid option in FOR_PS:"
		    " `%s' (Must be 'BEGIN' or 'END'\n\n",t);
	    exit(-1);
	}
	return 1;
    }
    /* store commands for PS or MIDI */
    if(ignore) {
	if(inspect) {
	    fprintf(stderr,"STORING LINE: '%s' whichbuf:(%d)\n",s,whichbuf);
	}
	if(whichbuf==PS_BUF) {
	    if(forpsbuflen<MAX_BUFLINES) {
		strncpy(forpsbuf[forpsbuflen],s,MAX_BUFLINELEN-1);
		forpsbuflen++;
	    }
	    else {
		fprintf(stderr,"\nERROR: forpsbuf[][]: Buffer Overfull\n\n");
		exit(-1);
	    }
	}
	else { /* MIDI_BUF */
	    if(formidibuflen<MAX_BUFLINES) {
		strncpy(formidibuf[formidibuflen],s,MAX_BUFLINELEN-1);
		formidibuflen++;
	    }
	    else {
		fprintf(stderr,"\nERROR: formidibuf[][]: Buffer Overfull\n\n");
		exit(-1);
	    }
	}
	if(inspect)
	    fprintf(stderr,"formidi:%d forps:%d\n",
		    formidibuflen,forpsbuflen);
	return 1;
    }

    if(sscanf(s,"echo:%s",t)!=0 && strlen(s)>=5) {
	if(textbuflen<(MAX_BUFLINES-1))
	    strncpy(textbuf[textbuflen++],s+5,MAX_BUFLINELEN-1);
	else
	    fprintf(stderr,"\nERROR: TEXT BUFFER OVERFULL\n\n");
	return 1;
    }

    if((c=sscanf(s,"commentmark: %s %s",comment[0],comment[1]))>0) {
	if(c==1)
	    comment[1][0]='\0';
	return 1;
    }

    /* strip comments */

    p=strstr(s,comment[0]); if(strlen(comment[0]) && p!=0) *p = '\0';
    p=strstr(s,comment[1]); if(strlen(comment[1]) && p!=0) *p = '\0';

    if(sscanf(s,"centre: %d",&centro)>0) {
	return 1;
    }

    if(sscanf(s,"center: %d",&centro)>0) {
	return 1;
    }

    if(sscanf(s,"name: %s",sname)>0) {
	strncpy(sname, s+5, MAX_NAME_LENGHT);
	clean_string(sname);
	return 1;
    }
    if(sscanf(s,"mod: %d",&modulo)>0 || sscanf(s,"size: %d",&modulo)>0) {
	if(modulo>0)
	    force_modulo=1;
	else if(modulo==0)
	    force_modulo=0;
	return 1;
    }
    if((c=sscanf(s,"range: %d %d",&ambito[0],&ambito[1]))>0) {
	if(c<2)
	    ambito[1]=ambito[0]+127;
	return 1;
    }
    if(sscanf(s,"transp: %d", &transp)>0)
	return 1;
    if(sscanf(s,"key: %d", &keymodulation)>0)
	return 1;

    c = SAG_DEFAULT_OCT;
    d = 0;
    if((x=sscanf(s,"sagittaloct: %d %d", &sagoctave[0], &sagoctave[1]))>0) {
	if(sagoctave[0] > 9 || sagoctave[0] < -9) {
	    fprintf(stderr,"\nERROR: Bad value for `sagittaloct:' (%d). "
		    "Resetting to %d.\n\n",
		    sagoctave[0],c);
	    sagoctave[0]=c;
	}
	if(x>1) {
	    if(sagoctave[1] > 9 || sagoctave[1] < -9) {
		fprintf(stderr,"\nERROR: Bad value for `sagittaloct:' (%d). "
			"Resetting to %d.\n\n",
			sagoctave[1], d);
		sagoctave[1]= d;
	    }
	}
	else
	    sagoctave[1]= d;
	return 1;
    }
    /* TODO: FIXME! IMPROVE THIS CODE! */
    if((c=sscanf(s,"alias: %d %d %d %d %d %d %d %d %d %d",
		 &indalias[0],&indalias[1],
		 &indalias[2],&indalias[3],
		 &indalias[4],&indalias[5],
		 &indalias[6],&indalias[7],
		 &indalias[8],&indalias[9] ))>0) {
	while(c < MAX_ALIAS) {
	    indalias[c]=-1; /*discard*/
	    c++;
	}

	return 1;
    }
    if((c=sscanf(s,"aliasabc: %d %d %d %d %d %d %d %d %d %d",
		 &abcalias[0],&abcalias[1],
		 &abcalias[2],&abcalias[3],
		 &abcalias[4],&abcalias[5],
		 &abcalias[6],&abcalias[7],
		 &abcalias[8],&abcalias[9] ))>0) {
	while(c < MAX_ALIAS) {
	    abcalias[c]=-1; /*discard*/
	    c++;
	}
	return 1;
    }
    if((c=sscanf(s,"aliassagittal: %d %d %d %d %d %d %d %d %d %d",
		 &sagittalias[0],&sagittalias[1],
		 &sagittalias[2],&sagittalias[3],
		 &sagittalias[4],&sagittalias[5],
		 &sagittalias[6],&sagittalias[7],
		 &sagittalias[8],&sagittalias[9] ))>0) {
	while(c < MAX_ALIAS) {
	    sagittalias[c]=-1; /*discard*/
	    c++;
	}
	return 1;
    }
    if(sscanf(s,"alias: %s",t)>0 &&
       (!strcmp(t,"all") ||
	!strcmp(t,"*"))) {
	for(c=0;c<MAX_ALIAS;c++)
	    indalias[c]=c;
	return 1;
    }
    if(sscanf(s,"aliasabc: %s",t)>0 &&
       (!strcmp(t,"all") ||
	!strcmp(t,"*"))) {
	for(c=0;c<MAX_ALIAS;c++)
	    indalias[c]=abcalias[c]=c;
	return 1;
    }
    if(sscanf(s,"aliassagittal: %s",t)>0 &&
       (!strcmp(t,"all") ||
	!strcmp(t,"*"))) {
	for(c=0;c<MAX_ALIAS;c++)
	    indalias[c]=sagittalias[c]=c;
	return 1;
    }
    if(sscanf(s,"up: %s",up)>0) {
	oitavanum=0;
	return 1;
    }
    if(sscanf(s,"down: %s",down)>0) {
	oitavanum=0;
	return 1;
    }
    if(sscanf(s,"updown: %s %s",up,down)>0) {
	oitavanum=0;
	return 1;
    }
    if(sscanf(s,"fixupdown: %d",&fix_updown)>0)
	return 1;
    if(sscanf(s,"modnum: %d",&oitavanum)>0)
	return 1;
    if((x=sscanf(s,"modnumsep: %d %c",&oitavasep,&oitavasepchar))>0) {
	if(x==1)
	    oitavasepchar=OCTSEP; /* default */
	return 1;
    }
    if(sscanf(s,"undef: %d",&undef)>0)
	return 1;
    if(sscanf(s,"chromatic: %d",&chroma)>0) {
	if(chroma) {
	    outputmode=CHROMATIC;
	    microtonal=0;
	    //denominator=0;
	    ///scl=-1;
	}
	///scl=-1; // TODO: check; is this proper?
	return 1;
    }
    if(sscanf(s,"diatonic: %d",&chroma)>0) {
	chroma=(!chroma);
	if(chroma==0) {
	    outputmode=DIATONIC;
	    //denominator=0;
	    microtonal=0;
	    ///scl=-1;
	}
	else
	    ;
	////microtonal=0; //denominator=0; /// TODO: check; is this proper?
	///scl= -1;//denominator=0; /// TODO: check; is this proper?
	return 1;
    }
    if(sscanf(s,"microtonal: %d",&microtonal)>0) {
	if(microtonal) {
	    outputmode=MICROTONAL;
	    chroma=1; // TODO: check
	}
	return 1;
    }
    if((c=sscanf(s,"replace: %d",&replace))>0) {
	if(replace>=0) {
	    chroma=1;
	    outputmode=LITERAL;
	}
	replace_abc=0;
	return 1;
    }

    if((c=sscanf(s,"replaceabc: %d %s",&replace,t))>0) {
	if(replace>=0) {
	    chroma=1;
	    replace_abc=1;
	    replace_updown=1;
	    outputmode=LITERAL;
	}
	if(c==2 && !strcmp(t,"noaccid"))
	    abcnoaccid=1;
	else
	    abcnoaccid=0;
	return 1;
    }
    if((c=sscanf(s,"replacesagittal: %d %s",&replace,t))>0) {
	if(replace>=0) {
	    chroma=1;
	    replace_sagittal=1;
	    replace_updown=1;
	    outputmode=LITERAL;
	}
	if(c==2) {
	    if(!strcmp(t,"noaccid"))
		sagittalnoaccid=1;
	    else
		sagittalnoaccid=0;
	    if(!strcmp(t,"visibleaccid"))
		sagittalvisibleaccid=1;
	    else
		sagittalvisibleaccid=0;
	}
	return 1;
    }
    if(sscanf(s,"replaceupdn: %d",&replace_updown)>0) {
	return 1;
    }
    if((c=sscanf(s,"replacenominal: %d %s",&replace_nominal,t))>0) {
	if(replace_nominal<0)
	    replace_nominal=0;
	if(c>1) {
	    if((pointer = strchr(abcnominals,t[0]))!=NULL) {
		nominal_offset = pointer - abcnominals;
	    }
	    pointer = t;
	    while(*pointer!='\0') {
		if(*pointer=='\'' || *pointer=='`') /* octave up */
		    nominal_offset+=7;
		else if(*pointer==',') /* octave down */
		    nominal_offset-=7;
		pointer++;
	    }
	}
	/////////
	fprintf(stderr,"\n>>> %d %d \n\n",replace_nominal,nominal_offset);
	/////////
	return 1;
    }

    if(sscanf(s,"scl: %d",&scl)>0) {
	if(scl>=0) {
	    microtonal=1;
	    /// replace=-1;    /// ?????
	    outputmode=MICROTONAL;
	}
	else {
	    microtonal=0;
	    chroma=1;
	}
	return 1;
    }
    if(sscanf(s,"sclmod: %s",t)>0) {
	sclmodulo=scl2cents(t);
	return 1;
    }
    if((x=sscanf(s,"equaltemp: %d %s",&modulo,t)) > 0) {
	if(modulo>0)
	    force_modulo=1;
	else if(modulo==0) {
	    force_modulo=0;
	}
	if(x>1)
	    sclmodulo=scl2cents(t);
	else
	    sclmodulo = 1200.0;
	equalt = 1;
	lineart = 0;
	scl = 1000;
	microtonal=1;
	outputmode=MICROTONAL;
	
	return 1;
    }

    t[0]=u[0]='\0';

    if((x=sscanf(s,"lineartemp:%d %s %s %d",&modulo, t,u, &lineart)) > 0) {
	// TODO: check this.
	generator = interv2cents(3.0/2.0);
	force_modulo=0;
	sclmodulo= 1200.0;
	microtonal=1;
	outputmode=MICROTONAL;
	
	if(modulo>0)
	    force_modulo=1;
	else if(modulo==0) {
	    force_modulo=0;
	}

	/* TODO: set default values here */
	if(x>1) {
	    sclmodulo=scl2cents(t);
	}
	else
	    sclmodulo=1200.0;

	if(x>2)
	    generator=scl2cents(u);
	else
	    generator=interv2cents(3.0/2.0);

	if(x>3) {
	    if( lineart==0 )
		lineart= -1; /* count downwards: if 0, represent it as
				negative */
	}
	else
	    lineart = (-1); /* as default */
	equalt = 0;
	scl = 1000;
	return 1;
    }
    t[0]='\0';
    if(sscanf(s,"sagittalapprox: %s",t)>0) {
	sagittal2scale = parse_sagittalapprox(s+15);
	return 1;
    }
    if(sscanf(s,"testfile: %s",testfilename)>0) {
	strncpy(testfilename, s+9, MAX_NAME_LENGHT);
	clean_string(testfilename);
	generate_testfile=1;
	return 1;
    }
    if(sscanf(s,"inputscl: %s",scala_input_file)>0) {
	strncpy(scala_input_file, s+9, MAX_NAME_LENGHT);
	clean_string(scala_input_file);
	inputscala= 1;
	microtonal=1;
	outputmode=MICROTONAL;
	scl=10000;
	return 1;
    }
    if(sscanf(s,"den: %d",&denominator)>0) {
	if(denominator<=0) {
	    denominator=16;
	    fprintf(stderr,
		    "\nERROR: INVALID `den:' RESET TO %d\n",
		    denominator);
	}
	if(abcm2ps_den && denominator>256) {
	    denominator=256;
	    fprintf(stderr,
		    "\nERROR: INVALID `den:' (USED WITH `abcm2ps:') RESET TO %d\n",
		    denominator);
	}
	return 1;
    }
    t[0]='\0';

    if((x=sscanf(s,"basefreq: %lf %d %s",&basefreq,&basepitch,t))>0) {
	/*	    if(x==3 && !strcmp(t,"+"))
		    basepluscentro=1;
		    else
		    basepluscentro=0; */
	if(x<2)
	    basepitch = 0;
	return 1;
    }

    if(sscanf(s,"afreq: %lf",&a_freq)>0)
	return 1;
    if((c=sscanf(s,"sagittalbasefreq: %lf %s",&sagittal_basefreq,t))>0) {
	if(c==1)
	    strncpy(sagittal_base,"A",1);
	else if (c>1) {
	    if(strlen(t)>1 || t[0]<'A' || t[0]>'G') {
		fprintf(stderr,
			"\nERROR: sagittalbasefreq: Invalid nominal '%s'\n"
			"       Should be an uppercase letter from A to G.\n\n",t);
		exit(-1);
	    }
	    else
		strncpy(sagittal_base,t,1);
	}
	return 1;
    }
    if(sscanf(s,"outfreq: %d",&outputfreq)>0)
	return 1;
    if(sscanf(s,"timidity: %s",timidity_file)>0) {
	strncpy(timidity_file, s+9, MAX_NAME_LENGHT);
	clean_string(timidity_file);
	if(!strcmp(timidity_file,"0")) {
	    timidity=0;
	    return 1;
	}
	timidity=1;
	chroma=1;
	microtonal=0;
	abcm2ps_den=0; /* incompatible */
	/* TODO: accept spaces in file name */
	/*strcpy(timidity_file,s+9);*/
	outputmode=CHROMATIC; ////////////// is this correct????
	return 1;
    }
    if(sscanf(s,"snt: %s",snt_file)>0) {
	strncpy(snt_file, s+4, MAX_NAME_LENGHT);
	clean_string(snt_file);
	if(!strcmp(snt_file,"0")) {
	    snt=0;
	    return 1;
	}
	outputmode=CHROMATIC;
	snt=1;
	chroma=1;
	microtonal=0;
	abcm2ps_den=0; /* incompatible */
	/* TODO: accept spaces in file name */
	/*strcpy(timidity_file,s+9);*/
	return 1;
    }
    if(sscanf(s,"scala: %s",scala_file)>0) {
	strncpy(scala_file, s+6, MAX_NAME_LENGHT);
	clean_string(scala_file);
	if(!strcmp(scala_file,"0")) {
	    outscala=0;
	    return 1;
	}
	outscala=1;
	///microtonal=0;
	////outputmode=MICROTONAL;  //// ????
	/* TODO: accept spaces in file name */
	/*strcpy(timidity_file,s+9);*/
	return 1;
    }
    if(sscanf(s,"scalakbm: %s",scalakbm_file)>0) {
	strncpy(scalakbm_file, s+9, MAX_NAME_LENGHT);
	clean_string(scalakbm_file);
	if(!strcmp(scalakbm_file,"0")) {
	    outscalakbm=0;
	    return 1;
	}
	outscalakbm=1;
	chroma=1;
	microtonal=0;
	outputmode=CHROMATIC;
	/* TODO: accept spaces in file name */
	/*strcpy(timidity_file,s+9);*/
	return 1;
    }
    if(sscanf(s,"abcm2ps: %d",&abcm2ps_den)>0) {
	if(abcm2ps_den>256 || abcm2ps_den<0) {
	    abcm2ps_den=1;
	    fprintf(stderr,
		    "\nERROR: INVALID `abcm2ps:' RESET TO %d\n",
		    abcm2ps_den);
	}
	if(abcm2ps_den!=0) {
	    outputmode=MICROTONAL;
	    microtonal=1;
	    if(denominator>256 || denominator<0) {
		denominator=256;
		/*    fprintf(stderr,
		      "\nERROR: `abcm2ps:' REQUIRES `den:' UP TO 256. RESET TO %d.\n",
		      denominator); */
	    }
	}
	return 1;
    }
    if(sscanf(s,"reduction: %d",&abcm2ps_reduction)>0)
	return 1;
    if(sscanf(s,"psacc: %d",&abcm2ps_psacc)>0)
	return 1;
    if(sscanf(s,"dumbacc: %d",&accid_dmacros)>0)
	return 1;
    if(sscanf(s,"defacc: %s %s",t,u)==2) {
	if(naccid>=MAX_ACCID-1)
	    fprintf(stderr,"\nERROR: defacc BUFFER FULL\n");
	else {
	    strcpy(accid_def[naccid][0],t);
	    strcpy(accid_def[naccid][1],u);
	    naccid++;
	}
	return 1;
    }

    if(sscanf(s,"procsag: %d",&procsag)>0)
	return 1;

    if(sscanf(s,"procmac: %d",&procmac)>0)
	return 1;

    if(sscanf(s,"skip: %s",skipstring)>0)
	return 1;

    if((x=sscanf(s,"delim: %s %s",leftd,rightd))>0) {
	if(!strcmp(leftd,"none"))
	    leftd[0]=rightd[0]='\0';
	else if(x==1)
	    rightd[0]='\0';
	return 1;
    }

    return 0; /* not identified */
}

int parse_sagittalapprox(char *s)
{
    int x, n;
    char t[BUFSIZ];
    char u[BUFSIZ];

    //// fprintf(stderr,"\nparse_sagittalapprox:\n\"%s\"\n",s);

    while(isspace(*s) && *s!='\0') /* remove leading spaces */
	s++;
    while((n=strlen(s)) > 0) { /* remove trailing spaces */
	if(isspace(s[n-1]))
	    s[n-1]='\0';
	else
	    break;
    }

    if((x=sscanf(s,"%s %s",t,u)) > 0) {
	if(strchr(t,'.') || strchr(t,'/')) { /* fifth size */
	    sag_generator=scl2cents(t);
	    sag_lineart=1;
	    sag_equalt=0;
	}
	else if (sscanf(t,"%d",&sag_modulo)) { /* equaltemp */
	    sag_equalt=1;
	    sag_lineart=0;
	}
	if(x>1) /* octave size */
	    sag_sclmodulo=scl2cents(u);
	else
	    sag_sclmodulo=1200.0;
    }
    else {
	fprintf(stderr,"\nERROR: Failed parse_sagittalapprox():\n\"%s\"\n",s);
	return 0;
    }
    /* ////////////
    fprintf(stderr,"sag_lineart: %d\n",sag_lineart);
    fprintf(stderr,"sag_equalt: %d\n",sag_equalt);
    fprintf(stderr,"sag_modulo: %d\n",sag_modulo);
    fprintf(stderr,"sag_sclmodulo: %lf\n",sag_sclmodulo);
    fprintf(stderr,"sag_generator: %lf\n",sag_generator);
    */ ///////////
    return 1;
}

int get_proc_command(char *s, FILE *fin, FILE *fout)
{
    char t[BUFSIZ]; /*u[BUFSIZ];*/
    char *p;

    /*    int c, x; */
    

    /* discard line with only spaces */
    if(espacos(s))
	return 0;

    /* remove blank spaces */
    strncpy(t,s,BUFSIZ-1);
    p=t;
    //    while(isspace(*p) && *p!='\0')
    while(isspace(*p) && *p!='\0')
	p++;
    strncpy(s,p,BUFSIZ-1);

    /* discard EOL characters */
    //    p=strchr(s,'\n'); if(p!=0) *p = '\0'; /* descarta '\n' (DOS) */

    p=strchr(s,13); if(p!=0) *p = '\0'; /* descarta eol */
    p=strchr(s,10); if(p!=0) *p = '\0'; /* descarta eol */

    t[0]='\0';
    if(sscanf(s,"include: %s",t)>0) {
	strcpy(t, s+8);
	clean_string(t);
	fprintf(stderr,"\n\n%s\n\n",t); ////
	includefile(fout,t);
	return 1;
    }

    ////fprintf(stderr,"s=`%s'\n",s);
    return 0; /* unknown */
}

int prepare()
{

    int i,n,c;
    double sc[MAX_SCALE_SIZE];
    double f;
    char s[BUFSIZ]; /* equal s[] size used in get_option()!!! */

    /* interpret FOR_PS:/FOR_MIDI: stores commands */
    if(forpsbuflen && processfor==FOR_PS) {
	for (i=0; i<forpsbuflen; i++) {
	    strcpy(s,forpsbuf[i]);
	    if(strlen(s) && !get_option(s))
		fprintf(stderr,
			"\nERROR: invalid option %s in `FOR_PS:' block\n\n",s);
	}
    }
    if(formidibuflen && processfor==FOR_MIDI) {
	for (i=0; i<formidibuflen; i++) {
	    strcpy(s,formidibuf[i]);
	    if(strlen(s) && !get_option(s))
		fprintf(stderr,
			"\nERROR: invalid option %s in `FOR_MIDI:' block\n\n",s);
	}
    }

    //////
    /// TODO: check/improve this!!!
    if(outputmode!=LITERAL)
	replace= -1;
    if(outputmode==MICROTONAL && scl<=-1)
	outputmode=CHROMATIC;

    if(embed_sagittal_fmt) {
	if (procsag && processfor!=FOR_PS) {
	    if(!easysag)
		fprintf(stderr,
			"WARNING: Option `-e' only works when "
			"preprocessing Sagittal for PostScript (-P).\n");
	    embed_sagittal_fmt=0; /* Don't warn again... */
	}
    }

    if(outputtofile)
	if(strcmp(outfile,abpfile)==0 || strcmp(outfile,infile)==0 ) {
	    fprintf(stderr,
		    "\nERROR: INPUT AND OUTPUT FILES ARE THE SAME: `%s'\n\n",
		    outfile);
	    exit(-1);
	}
		
    if(sclmodulo==0.0) {
	fprintf(stderr,
		"\nERROR: prepare(): "
		"Cannot continue: sclmodulo==0.0 (1/1)\n\n");
	exit(-1);
    }
    if(modulo<1) {
	fprintf(stderr,"\nERROR: prepare(): Cannot continue: modulo<=0\n\n");
	exit(-1);
    }
    if(modulo>MAX_SCALE_SIZE) {
	fprintf(stderr,
		"\nERROR: prepare(): "
		"Cannot continue: modulo>%d (MAX_SCALE_SIZE)\n\n",
		MAX_SCALE_SIZE);
	exit(-1);
    }

    /* TODO: check: is here the correct place for this? */

    if(equalt) {
	equaltempc(modulo,(sclmodulo),scale);
	scl=10000;
    }
    else if(lineart) {
	lineartempc(modulo,(sclmodulo),
		    (generator),
		    (lineart<0 ? 0 : lineart),
		    scale);
	scl=10000;
    }

    if(inputscala) {
	n=readscalafile(scala_input_file);
	if(force_modulo==0)
	    modulo=n;
	scl=10000;
    }

    amin = ambito[0] + centro;
    amax = ambito[1] + centro;

    /// if(chroma==0) { /* adjust for diatonic pitches only */
    if(outputmode==DIATONIC) { /* adjust for diatonic pitches only */
	while(!diatonic(centro)) {/* TODO: put warning when changed! */
	    centro++;
	}
	amin=centro;
	for(c=0; c<abs(ambito[0]); c++) {
	    amin--;
	    while(!diatonic(amin))
		amin--;

	}
	amax=centro;
	for(c=0; c<abs(ambito[1]); c++) {
	    amax++;
	    while(!diatonic(amax))
		amax++;

	}
    }

    /* apply keymodulation to the scale */
    keymodulation = mod(keymodulation,modulo);
    if(keymodulation) {
	for(i=0; i<modulo; i++) {
	    sc[i]=scale[mod(i+keymodulation,modulo)];
	    if(sc[i]>=sclmodulo)
		sc[i]-=sclmodulo;
	    if(sc[i]<0.0)
		sc[i]+=sclmodulo;
	}
	f=scale[keymodulation];
	for(i=0; i<modulo; i++) {
	    scale[i]=sc[i]-f;
	    if(scale[i]>=sclmodulo)
		scale[i]-=sclmodulo;
	    if(scale[i]<0.0)
		scale[i]+=sclmodulo;
	}
    }

    if(scl>=0)
	set_basefreq();

    /* TODO: Fix to avoid wrong warnings;
     * Yet better: move to inside the code which generate macros
     */
    /*
      if(amin-centro<0 || amax>127) {
      fprintf(stderr,
      "WARNING: MIDI range from %d (%s) to %d (%s)\n",
      amin,(abc(s,amin),s),amax,(abc(t,amax),t));
      }
    */

    prepare_sagittal();

    if(replace>=0 && scl<0) /// ????
	centro=0;

    if(inspect && scl>=0) {
	/* show the scale on the screen */
	fprintf(stderr,"\nSCALE DATA (size=%d):\n\n",modulo);
	for(i=1;i<modulo;i++)
	    fprintf(stderr,"%10.3lf\n",scale[i]);
	fprintf(stderr,"%10.3lf\n",sclmodulo);
	fprintf(stderr,"\n");
    }

    return 1;
}

int output(FILE *fp)
{
    int i,j,p,c,x,n, count=0;
    char s[BUFSIZ];
    char t[BUFSIZ];

    prepare();

    fprintf(fp,"#\n# File automatically generated by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n#\n");

    if(scl<0)
    	p=ambito[0]+centro;
    else
       	p=ambito[0];

    if(replace<0)
	p-=transp;
    else
	outputfreq = 0; /* don't output freq in literal mode */


    syncalias();

    if(textbuflen) { /* echo must be output first to allows %%format file  */
	fprintf(fp,"#\n# ===================={   START OF ECHO TEXT\n");
	for(i=0; i<textbuflen; i++) { /* output echo: lines of text */
	    fprintf(fp,"%s",textbuf[i]);
	    if(textbuf[i][strlen(textbuf[i])-1]!='\n')
		fprintf(fp,"\n"); /* print newline if missing (command line) */
	}
	fprintf(fp,"# ====================}    END OF ECHO TEXT\n#\n");
    }

    c=0;
    fprintf(fp,"#\n# ===================={  START OF DEFINITIONS\n");
    if(sname[0]!='\0') {
	fprintf(fp,"#\n#define %s \"\"\n#\n",sname);
	strcpy( macrodef[nmacrodefs][0], sname );
	strcpy( macrodef[nmacrodefs][1], "" );
	nmacrodefs++;
	count++;
    }
    if(accid_dmacros) {
	if(accid_dmacros>0) {	/* void macros */
	    if(inputabpfile==0) { /* output accidental macros */
		fprintf(fp,"#define %s^%s \"\"\n",leftd,rightd); /*sharp*/
		fprintf(fp,"#define %s=%s \"\"\n",leftd,rightd); /*natural*/
		fprintf(fp,"#define %s_%s \"\"\n",leftd,rightd); /*flat*/
		fprintf(fp,"#define %s^^%s \"\"\n",leftd,rightd); /*dsharp*/
		fprintf(fp,"#define %s__%s \"\"\n",leftd,rightd); /*dflat*/
		/* user defined accidentals */
		for(i=0;i<naccid;i++)
		    fprintf(fp,"#define %s%s%s \"\"\n",
			    leftd,accid_def[i][0],rightd);
	    }
	    //else
	    {
		/* store accidentals macros for built-in preprocessor */
		strcpy(macrodef[nmacrodefs][0],"^"); /*sharp*/
		macrodef[nmacrodefs++][1][0] = '\0';

		strcpy(macrodef[nmacrodefs][0],"="); /*natural*/
		macrodef[nmacrodefs++][1][0] = '\0';

		strcpy(macrodef[nmacrodefs][0],"_"); /*flat*/
		macrodef[nmacrodefs++][1][0] = '\0';

		strcpy(macrodef[nmacrodefs][0],"^^"); /*dsharp*/
		macrodef[nmacrodefs++][1][0] = '\0';

		strcpy(macrodef[nmacrodefs][0],"__"); /*dflat*/
		macrodef[nmacrodefs++][1][0] = '\0';

		/* user defined accidentals */
		/* TODO: check: should undefine accid macros ? */
		for(i=0;i<naccid;i++) {
		    strcpy( macrodef[nmacrodefs][0], accid_def[i][0] );
		    strcpy( macrodef[nmacrodefs][1], accid_def[i][1] );
			nmacrodefs++;
		}
	    }
	}
	else { /* accid_dmacros < 0 --- does define accidental macros */
	    if(inputabpfile==0) {
		fprintf(fp,"#define %s^%s ^\n",leftd,rightd);
		fprintf(fp,"#define %s=%s =\n",leftd,rightd);
		fprintf(fp,"#define %s_%s _\n",leftd,rightd);
		fprintf(fp,"#define %s^^%s ^^\n",leftd,rightd);
		fprintf(fp,"#define %s__%s __\n",leftd,rightd);
		for(i=0;i<naccid;i++)
		    fprintf(fp,"#define %s%s%s %s\n",
			    leftd,accid_def[i][0],rightd,accid_def[i][1]);
	    }
	    // else
	    { 
		/* store user defined accidentals for preprocessor */
		/* store accidentals macros for built-in preprocessor */
		strcpy(macrodef[nmacrodefs][0],"^"); /*sharp*/
		strcpy(macrodef[nmacrodefs++][1],"^"); /*sharp*/

		strcpy(macrodef[nmacrodefs][0],"="); /*natural*/
		strcpy(macrodef[nmacrodefs++][1],"="); /*natural*/

		strcpy(macrodef[nmacrodefs][0],"_"); /*flat*/
		strcpy(macrodef[nmacrodefs++][1],"_"); /*flat*/

		strcpy(macrodef[nmacrodefs][0],"^^"); /*dsharp*/
		strcpy(macrodef[nmacrodefs++][1],"^^"); /*dsharp*/

		strcpy(macrodef[nmacrodefs][0],"__"); /*dflat*/
		strcpy(macrodef[nmacrodefs++][1],"__"); /*dflat*/

		for(i=0;i<naccid;i++) {
		    strcpy( macrodef[nmacrodefs][0], accid_def[i][0] );
		    strcpy( macrodef[nmacrodefs][1], accid_def[i][1] );
		    nmacrodefs++;
		}
	    }
	}
	count += 5 + naccid;
    }
    
    for(i=amin;i<=amax;i++) { /* FOR EACH PITCH */
	////if(!chroma && !diatonic(i))
	if(outputmode==DIATONIC && !diatonic(i))
	    continue; /* skip chromatic pitches when not chroma */
	n=alias(s,p,100);
	for(j=0;j<n;j++) { /* FOR EACH ALIAS */
	    for(x=0;x<n;x++) /*test whether j is index for a selected column*/
		if(j==indalias[x])
		    break;
	    if (x<n) {
		if(replace>=0) {
		    alias(s,p+transp,replace);
		    if(s[0]=='\0') {
			alias(s,p,j);
			fprintf(stderr,
				"\nERROR! MISSING REPLACEMENT FOR %s%s%s!\n\n",
				leftd,s,rightd);
			continue; /* skip void strings in replacements */
		    }
		}

		alias(s,p,j);
		
		if(s[0]=='\0')
		    continue; /* skip void strings in aliases */

		for(x=0;x<MAX_ALIAS;x++) /* test whether j is index
					    for an aliasabc column */
		    if(abcalias[x]==j)
			fixabc(s);

		/* ********** TODO: centralize output (use a buffer) */
		if(inputabpfile==0)
		    fprintf(fp,"#define %s%s%s ",leftd,s,rightd);/*MACRO_NAME*/
		//else
		/* store macro name for built-in preprocessor */
		strncpy(macrodef[nmacrodefs][0],s,64);

		if(!undef) {
		    if(replace>=0) { /* literal mode */
			if(replace_updown==0)
			    alias(s,mod(p+transp,modulo),replace);
			else {
			    alias(s,p+transp,replace);
			    if(replace_abc) {
				fixabc(s); /* output valid ABC pitches */
				if(abcnoaccid)
				    remove_accid(s);
			    }
			    else if(replace_sagittal) {
				/* TODO: sagittal replacement without octave? */
				/* TODO: option to don't put delimiters? need or not? */
				if(sagittalnoaccid) {
				    if(s[0]!=DONTSHOW)
					sprintf(t,"%s%c%s%s",leftd,DONTSHOW,s,rightd);
				    else
					sprintf(t,"%s%s%s",leftd,s,rightd);
				}
				else if(sagittalvisibleaccid) {
				    if(s[0]==DONTSHOW)
					sprintf(t,"%s%s%s",leftd,s+1,rightd);
				    else
					sprintf(t,"%s%s%s",leftd,s,rightd);
				}
				else
				    sprintf(t,"%s%s%s",leftd,s,rightd);
				strncpy(s,t,BUFSIZ-1);
			    }
			}
		    }
		    else if(microtonal) {
			/* scl>=0 && !timidity && !snt) { microtonal mode*/
			alias(s,p+transp,scl);
			microabc(s,p+transp); ////////
			///fprintf(stderr,"%s => %f Hz\n", s, abc2freq(s)); ////
		    }
		    else { /* diatonic or chromatic mode */
			//// fprintf(stderr,"Will call abc() with '%s'\n",s);
			abc(s,i,outputmode);
		    }

		    /* TODO: Check it!!! Strange use of `centro' */
		    if((timidity || snt) && scl>=0) {
			/* freq=pitch2freq(p); */
			/*  if(snt) */
			/* freq=pitch2freq(p+centro); */
			freq=pitch2freq(p+centro+transp);
		    }
		    else if(chroma && scl<0)
			///else if(outputmode==CHROMATIC)
			freq=midi2freq(p+transp,a_freq);

		    /* ********** TODO: centralize output (use a buffer) */
		    if(inputabpfile==0)
			fprintf(fp,"%s",s); /* MACRO_REPLACEMENT */
		    //else
		    /*store macro replacement for built-in
		      preprocessor*/
		    strncpy(macrodef[nmacrodefs++][1],s,64);
		    if(nmacrodefs>=MAX_DEFINITIONS) {
			fprintf(stderr,
				"\nERROR: DEFINITIONS OVERFULL (%d #defines)\n\n",
				nmacrodefs);
			exit(-1);
		    }

		}

		/* Output frequencies */
		if(scl>=0 && outputfreq && !undef && freq!=0.0) {
		    freq=pitch2freq(p+transp); ////  /* due do `centro' */
		    fprintf(fp,"\n#       %lf Hz\n# ",freq);
		    fprintf(stderr,"#       %lf Hz\n",freq);
		}
		
		/* ********** TODO: centralize output (use a buffer) */
		if(inputabpfile==0)
		    fprintf(fp,"\n");
		c++;
	    }
	}
	p++; /* increment p */
    }
    fprintf(fp,"# ====================}   END OF DEFINITIONS  ");
    count+=c;
    fprintf(fp,"(TOTAL: %d)\n#\n",count);

    if(scl>=0) {
	/*** this MUST stay ... ***/
	if(abcm2ps_den==4 && abcm2ps_psacc) { /* eighths of tone */
	    fprintf(fp,
		    "#\n# ===================={"
		    "   START OF abcm2ps 1/8-TONE ACCIDENTALS\n");
	    output_psacc(fp);
	    fprintf(fp,
		    "# ====================}"
		    "    END OF abcm2ps 1/8-TONE ACCIDENTALS\n#\n");
	}
	/*** ... before this!: ***/

	/// if(abcm2ps_den && !timidity && !snt && !undef) {
	if(abcm2ps_den && outputmode==MICROTONAL) {
	    if(procsag && processfor==FOR_MIDI && denominator<=256) {
		/* When preprocessing Sagittal music for MIDI and
		   `abcm2ps:', don't know in advance which accidentals
		   will occur, then store all the possible ones */
		for(i=1; i<=denominator; i++) {
		    storenumerator(-i);
		    storenumerator(i);
		}
	    }
	    fprintf(fp,
		    "#\n# ===================={"
		    "   START OF abcm2ps ACCIDENTALS\n");
	    i=print_ps_accid(fp);
	    fprintf(fp,
		    "# ====================}    END OF abcm2ps ACCIDENTALS ");
	    fprintf(fp," (TOTAL: %d)\n#\n",i);
	}
    }

    /*
      if(timidity && scl>=0)
      output_timidity();
      if(snt && scl>=0) // abc2midi Standard MIDI Tuning as snt commands
      output_snt(NULL); // NULL to send output to snt_file (not to .abh)
      if(outscala && scl>=0)
      output_scala();
      if(outscalakbm && scl>=0)
      output_scalakbm();
    */
    if( (scl<0) && (timidity||snt||outscala) ) {
	fprintf(stderr,"\nERROR: INDEFINED TUNING FOR:"
		"%s%s%s\n\n",
		timidity? " `timidity:'":"",
		snt? " `snt:'":"",
		outscala? " `scala:'":"");
    }
    else {
	if(timidity)
	    output_timidity();
	if(snt) // abc2midi Standard MIDI Tuning as snt commands
	    output_snt(NULL); // NULL to send output to snt_file (not to .abh)
	if(outscala)
	    output_scala();
    }
    if(outscalakbm)
	output_scalakbm();

    return 1;
}

int syncalias()
{
    int i,j,c;

    for(i=0; i<MAX_ALIAS; i++) { /* add new abc aliases */
	c=1;
	if(abcalias[i]<0)
	    break;
	for(j=0; j<MAX_ALIAS; j++) {
	    if((indalias[j]==abcalias[i])) {
		c=1; /* found */
		break;
	    }
	    if(indalias[j]<0) { /* end of valid alias */
		c=0; /* not found */
		break;
	    }
	}
	if(c==0 && j<MAX_ALIAS && c<MAX_ALIAS) { /* not found; add it */
	    indalias[j]=abcalias[i]; /* copy new alias */
	}
    }

    for(i=0; i<MAX_ALIAS; i++) { /* add new sagittal aliases */
	c=1;
	if(sagittalias[i]<0)
	    break;
	else
	    usingsagittalalias=1;
	for(j=0; j<MAX_ALIAS; j++) {
	    if((indalias[j]==sagittalias[i])) {
		c=1; /* found */
		break;
	    }
	    if(indalias[j]<0) { /* end of valid alias */
		c=0; /* not found */
		break;
	    }
	}
	if(c==0 && j<MAX_ALIAS && c<MAX_ALIAS) { /* not found; add it */
	    indalias[j]=sagittalias[i]; /* copy new alias */
	}
    }

    /* if no alias, select 1 */
    c=0;
    for(i=0;i<MAX_ALIAS;i++)
	if(indalias[i]>=0 && indalias[i]<MAX_ALIAS) {
	    c=1;
	    break;
	}
    if(c==0)
	indalias[0]=1;    
    return 1;
}

int alias(char *s, int n, int i)
{
    int oitava, cont, y, x, w;
    int issag, nom, j=0, nommod;
    char t[MAX_ALIAS][64], r[64];
    char up2[8], dn2[8], *p;
    const char nominalstr[] = "CDEFGAB";

    s[0]='\0';
    issag=0;

    if(scl<0)
	n-=centro;

    /* TODO: improve!!! */
    cont=sscanf(aliases[mod(n,modulo)],
		"%s %s %s %s %s %s %s %s %s %s",
		t[0], t[1], t[2], t[3], t[4],
		t[5], t[6], t[7], t[8], t[9]);

    for(y=0; y<cont; y++) /* ignore item "." */
	if(!strcmp(t[y],skipstring))
	    t[y][0]='\0';

    if(i>=0 && i<cont)
	strncat(s, t[i], BUFSIZ-1-strlen(s));
    else
	return cont;
    
    if(s[0]=='\0') /* ignored item */
	return cont;

    for(x=0;x<MAX_ALIAS;x++) /* test whether i is index
				for an aliassagittal column */
	if(sagittalias[x]==i) {
	    normalizesagittal(s); /* normalize sagittal pitch */
	    issag=1;
	}

    ///////////////////////// TEST
    nommod = replace_nominal;
    ///////////////////////// TEST

    if(i==replace && replace_nominal) {
	/* replace [n] with nominal */
	////	fprintf(stderr,"1) alias(): s=\"%s\"\n",s);
	r[0]='\0';
	if((p=strchr(s,'['))!=NULL) {
	    j=p-s;
	    if(j) {
		strncpy(r,s,j);
		r[j]='\0';
		j=strlen(r);
	    }
	    /*  fprintf(stderr,"%s %s\n",s,r); */
	    if(sscanf(p+1,"%d",&nom)>0) {
		r[j++] = nominalstr[mod(nominal_offset+
				      (nom+((n-mod(n,modulo))/modulo)*nommod),7)];
		if((p=strchr(s,']'))!=NULL) {
		    for (;*p!='\0';)
			r[j++]=*(++p);
		}
		else {
		    fprintf(stderr,"\nalias(): CANNOT CONVERT NOMINAL FOR `%s'\n\n",s);
		    exit(-1);
		}
		r[j]='\0';
	    }
	    else {
		fprintf(stderr,"\nalias(): CANNOT CONVERT NOMINAL FOR `%s'\n\n",s);
		exit(-1);
	    }
	}
	strcpy(s,r);
	j = nominal_offset+nom+((n-mod(n,modulo))/modulo)*nommod;
	j = (j-mod(j,7))/7;
	/*
	  while (j < 0) {
	  strcat(s,",");
	  j++;
	  }
	  while (j > 0) {
	  strcat(s,"`");
	  j--;
	  }
	*/
	////////
	n=j*modulo; //// Control octave modifiers.
	////////
    }

    if(i==replace && replace_sagittal) {
	    normalizesagittal(s); /* normalize sagittal pitch */
	    issag=1;
    }

    oitava=(n-mod(n,modulo))/modulo;

    if(oitavanum==0 || issag) { /* octave as string */
	
	if(issag) { /* temporarily change up/dn for Sagittal */
	    strncpy(up2,up,7);
	    strncpy(dn2,down,7);
	    up[0]=SAGOCTAVEUP;
	    down[0]=SAGOCTAVEDN;
	    up[1]=down[1]= '\0';
	}
	if(fix_updown) /* remove up/down redundances (pass 1) */
	    remove_bad_updown(s);
	if(oitava<=0)
	    for(;oitava<0;oitava+=(oitava>0?-1:1)) {
		if(strlen(s)>=64-1-strlen(down)) ////////////
		    break;
		strcat(s,down);
	    }
	else
	    for(;oitava>0;oitava+=(oitava>0?-1:1)) {
		if(strlen(s)>=64-1-strlen(up)) /////////////
		   break;
		strcat(s,up);
	    }
	if(fix_updown) /* remove up/down redundances (pass 2) */
	    remove_bad_updown(s);

	if(issag) { /* reset normal up/dn */
	    strncpy(up,up2,7);
	    strncpy(down,dn2,7);
	}

    }
    else if(oitava) { /* octave as number */
	//
	/* TODO: (?) treat octave number given in alias */
	//
	if(oitavasep) {
	    w = strlen(s);
	    if(w<64-2) {
		s[w] = oitavasepchar;
		s[w+1] = '\0';
	    }
	    //	    strcat(s," ");
	}
	sprintf(t[4],"%d",oitava);
	strncat(s,t[4],64-1-strlen(s));
	//	strcat(s,t[4]);
    }
    return cont;
}

int remove_bad_updown(char *s)
{
    char *c1, *c2;
    char *p, *u;
    int s1, s2;
    int i, pisup,x;

    for(x=0;x<5;x++) { /* awful hack to fix more pairs of bad updowns */
	pisup=0;
	s1=strlen(s);
	s2=strlen(up)+strlen(down);
	
	/* TODO: fix bad octave numbers??? */
	
	/* TODO: FIXME! By now, only removes first pair of up/down */
	if((c1=strstr(s,up))!=NULL && (c2=strstr(s,down))!=NULL) {
	if(c1<c2)
	    p=c1, u=c2, pisup=1;
	else
	    p=c2, u=c1, pisup=0;
	for(i=0;i<s2;i++)
	    *(p+i)=*(p+i+1);
	s[s1-s2]='\0';
	}
    }
    return 1;
}


/* convert pitch of scale into ABC chromatic representation */
int abc(char *s, int n, int typ) /* typ==DIATONIC omits accidental */
{
    int oitava, i, c=2;

    s[0]=s[1]=s[2]='\0';

    switch(mod(n,12)) {
    case  0: s[0]='='; s[1]='C'; break;
    case  2: s[0]='='; s[1]='D'; break;
    case  4: s[0]='='; s[1]='E'; break;
    case  5: s[0]='='; s[1]='F'; break;
    case  7: s[0]='='; s[1]='G'; break;
    case  9: s[0]='='; s[1]='A'; break;
    case 11: s[0]='='; s[1]='B'; break;
    case  1: s[0]='^'; s[1]='C'; break;
    case  3: s[0]='^'; s[1]='D'; break;
    case  6: s[0]='^'; s[1]='F'; break;
    case  8: s[0]='^'; s[1]='G'; break;
    case 10: s[0]='^'; s[1]='A'; break;
    default: 
	s[0]='\0';
    }

    oitava=(n-mod(n,12)-60)/12;

    if(oitava<0) {
	for(;oitava<0;oitava+=(oitava>0?-1:1)) {
		if(c>=64-1) {///////////
		    fprintf(stderr,
			    "\nERROR: abc(): Too many octave indicators."
			    " Bad `range:'? (FATAL)\n\n");
		    exit(-1); ///////////
		}
		s[c++]=',';
	}
    }
    else {
	if(oitava>0)
	    for(i=0; i<3; i++)
		s[i]=tolower(s[i]);
	for(oitava--;oitava>0;oitava+=(oitava>0?-1:1))
	    {
		if(c>=64-1) {///////////
		    fprintf(stderr,
			    "\nERROR: abc(): Too many octave indicators. Bad `range:'? (FATAL)\n\n");
		    exit(-1); ///////////
		}
		s[c++]='\'';
	    }
    }
    s[c++]='\0';

    /* if in diatonic mode, remove accidental (natural) */
    ///    if(!chroma || (scl>=0 && !timidity && !snt) || microtonal) {
    //    if( (outputmode==DIATONIC)
    //|| outputmode==MICROTONAL
       // || (scl>=0 && !timidity && !snt)
       // || microtonal // not the mode!
    if(typ==DIATONIC) {
	c=strlen(s);
	for(i=1;i<=c;i++) /* until '\0' */
	    s[i-1]=s[i];  /* to left */
    }
    
    return 0;
}

int fixabc(char *s) /* fix bad ABC pitches like C' C', c,, C'' */
{
    int c,i,n,oct,flag;
    char *p;
    char up0[8], down0[8];
    int maxlen=BUFSIZ-1;

    /* remove all blank spaces */
    while((p=strchr(s,' '))!=NULL) {
	while(*(p+1)!='\0') {
	    *p = *(p+1);
	    p++;
	}
	*p='\0';
    }

    /* fix octaves as numbers */

    /* find number after nominal */
    n=strlen(s);
    flag=0;
    for(i=0;i<n;i++) {
	c = s[i];
	if(tolower(c)>='a' && tolower(c)<='g')
	    flag=1;
	else if(flag) {
	    if(isdigit(c) || c=='-')
		break;
	}
    }
    oct=0;
    if(i<n) { /* there was a bad octave number -- get it! */
	sscanf(s+i,"%d",&oct);
    }
    s[i]='\0'; /* remove number */
    while(oct!=0) { /* put '' or ,, */
	if(strlen(s)>=maxlen-1)
	    break;
	if(oct>0) {
	    strcat(s,"'");
	    oct--;
	}
	else { /* oct<0 */
	    strcat(s,",");
	    oct++;
	}
    }

    /* temporarily change updown to ABC ' and , */
    strncpy(up0,up,7);
    strncpy(down0,down,7);
    strcpy(up,"'");
    strcpy(down,",");

    remove_bad_updown(s);

    /* reset updown back */
    strncpy(up,up0,7);
    strncpy(down,down0,7);

    for(c='A';c<='G';c++) {
	p=strchr(s,c);
	if(p==NULL)
	    continue;
	if(strchr(s,'\'')!=NULL) {
	    *p=tolower(c);
	    s[strlen(s)-1]='\0'; /* remove trailing ' */
	}
    }
    for(c='a';c<='g';c++) {
	p=strchr(s,c);
	if(p==NULL)
	    continue;
	if(strchr(s,',')!=NULL) {
	    *p=toupper(c);
	    s[strlen(s)-1]='\0'; /* remove trailing , */
	}
    }
    return 1;
}

int remove_accid(char *s) /* remove ABC accidental */
{
    int c;
    char t[BUFSIZ], *p;

    for(c='a'; c<='g'; c++) {
	p=strchr(s,c);
	if(p!=NULL) {
	    strcpy(t,p);
	    strcpy(s,t);
	    return 1;
	}
    }
    for(c='A'; c<='G'; c++) {
	p=strchr(s,c);
	if(p!=NULL) {
	    strcpy(t,p);
	    strcpy(s,t);
	    return 1;
	}
    }
    return 1;
}

double pitch2freq(int pitch)
{
    int imod, pc;
    /* #include <math.h>
       long double powl(long double, long double); */
    if(modulo==0) {
	fprintf(stderr,"\nERROR: pitch2freq() => modulo==0\n\n");
	exit(-1);
    }

    imod=(pitch-mod(pitch,modulo))/modulo; /* modulo instance number */
    pc=mod(pitch,modulo);                  /* pitch class */

    freq= (double)
	((long double) basefreq
	 * powl((long double)cents2interv(sclmodulo),(long double)imod)
	 * cents2interv(scale[pc]));

    return freq;
}

int set_basefreq()
{
    double f1, f2;
    // fprintf(stderr,"\nbasefreq: %.3lf Hz\n",basefreq);
    // fprintf(stderr,"basepitch: %d\n",basepitch);
    f1 = pitch2freq(basepitch);
    f2 = pitch2freq(0); /* basefreq */
    basefreq/=(f1/f2);
    // fprintf(stderr,"basefreq: %.3lf Hz\n",basefreq);
    // fprintf(stderr,"basepitch: %d\n",basepitch);
    basepitch=0;
    // fprintf(stderr,"basepitch: %d\n\n",basepitch);
    return 1;
}

int microabc(char *s, int p) /* from pitch of scale to ABC microtonal code */
{
    pitch2freq(p);
    if(abcm2ps_den)
	freq/=a_freq/440.0;
    else
	freq*=a_freq/440.0;
    return freq2abc(s,freq);
}

int freq2abc(char *s, double f) /* from frequency to ABC microtonal code */
{
    double midi,cents;
    int nearest,numerator;
    int n,d;
    char t[64];

    /// midi = freq2midi(f,a_freq); ///// Is this correct???
    midi = freq2midi(f,440.0); ///// Is this correct???

    nearest=nearestdiat(midi);
    cents=(midi-(double)nearest)*100.0;
    numerator=cents2frac(cents,denominator); /* ratio numerator */

    /* debug */
    if(abs(numerator)>denominator)
	fprintf(stderr,"\nWARNING: freq2abc: fix nearestdiat() !!!\n"
		"numerator=%d denominator=%d\n\n",numerator,denominator);

    /* numerator==denominator is not always reduced due to abc2midi
       idiosynchrasy (when using `snt:' AND pitchbends, accidentals
       should be based on the natural pitches (nominals), not in the
       nearest pitch in the chromatic scale); Control this with 1
       reduction:' */

    n=numerator;
    d=denominator;
    if(abcm2ps_reduction > 1) {
	if(abs(n)==d) {
	    d=1;
	    n=(n>0)?1:-1;
	}
	else if((abs(n)*2)==d
		&& (abcm2ps_den%2==0 || abcm2ps_reduction > 2 ) ) {
	    d=2;
	    n=(n>0)?1:-1;
	}
	else if(numerator)
	    storenumerator(numerator);
    }
    else if(numerator)
	storenumerator(numerator);

    if (n) /* microtonal accidental */
	{
	    if(d==1)
		sprintf(t,"%c",n>0?'^':'_');
	    else
		sprintf(t,"%c%d/%d",n>0?'^':'_', abs(n),d);
	}
    else
	sprintf(t,"="); /* natural sign */

    abc(s,nearest,DIATONIC); /* basic abc note name */

    ////fprintf(stderr,"%s %d\n",s,nearest); /////////

    strncat(t,s,64-1-strlen(t)); /////////////////
    ////strcat(t,s);
    strcpy(s,t);

    return 0;
}

double abc2freq(char *s)  /* Convert (microtonal) ABC into frequency */
{
    char *p, c, nominal;
    int accidental, num, den, oct, n;
    double f;
    int refA;

#define DFLAT   0
#define FLAT    1
#define NATURAL 2
#define SHARP   3
#define DSHARP  4
#define NONE    5

    nominal=' ';
    accidental=NONE;
    oct=0;

    /// refA = a_freq;  //// is this correct?
    refA = 440.0;
    f = 1.0;

    p=s;
    c=*p;
    if(c=='=' || c=='^' || c=='_') { /* abc accidental */
	if(c=='=')
	    accidental=NATURAL;
	else if (c=='^') {
	    if(*(p+1)=='^') {
		accidental=DSHARP;
		p++;
	    }
	    else
		accidental=SHARP;
	}
	else if (c=='_') {
	    if(*(p+1)=='_') {
		accidental=DFLAT;
		p++;
	    }
	    else
		accidental=FLAT;
	}
	p++;
	
	/* treat microtone accidentals */
	num=1;
	den=1;
	if(*(p)=='/' || isdigit(*(p))) {
	    if(*p=='/') {
		if((n=sscanf(p,"/%d",&den))==0)
		    den=2;
	    }
	    else if((n=sscanf(p,"%d/%d",&num,&den))) {
		if(n==1)
		    den=2;
	    }
	    else {
		num=1;
		den=1;
		///fprintf(fout,"[r:nomicrotone]");
	    }
	    while(isdigit(*p) || *p=='/')
		p++;
	}
	c=*p;
	/* ///  fprintf(fout,"[r:acc=%d,num=%d,den=%d *p=%c]",
	   pitch.accidental,
	   pitch.num,
	   pitch.den,c); */
    }
    ////fprintf(fout,"[r:%c]",*p);
    if(tolower(c) >= 'a' && tolower(c) <= 'g') { /* abc pitch */
	nominal=c;
	p++;
	/* try get octave */;
	while((c=*(p))=='\'') {
	    oct++;
	    p++;
	}
	while((c=*(p))==',') {
	    oct--;
	    p++;
	}
	///c=*p;
    }

    if(islower(nominal)) {
	nominal = toupper(nominal);
	oct++;
    }

    switch (nominal) {
    case 'C':
	f = midi2freq(60.0,refA);
	break;
    case 'D':
	f = midi2freq(62.0,refA);
	break;
    case 'E':
	f = midi2freq(64.0,refA);
	break;
    case 'F':
	f = midi2freq(65.0,refA);
	break;
    case 'G':
	f = midi2freq(67.0,refA);
	break;
    case 'A':
	f = midi2freq(69.0,refA);
	break;
    case 'B':
	f = midi2freq(71.0,refA);
	break;
    default:
	fprintf(stderr,"\nERROR: abc2freq() cannot convert '%s'\n\n",s);
	return 1.0;
    }

    if(accidental==DFLAT) {
	f /= cents2interv(200.0);
    }
    else if(accidental==DSHARP) {
	f *= cents2interv(200.0);
    }
    else if(accidental==FLAT) {
	f /= cents2interv(100.0*num/den); // float!
    }
    else if(accidental==SHARP) {
	f *= cents2interv(100.0*num/den); // float!
    }

    f *= powl(2.0,oct);

    return f;
}

int storenumerator(int n)
{
    int i;
    for(i=0; i<nnumerators && i<MAX_NUMERATORS; i++) {
	if(numerators[i]==n)
	    return 1;
    }
    if (nnumerators < MAX_NUMERATORS) {
	numerators[nnumerators++] = n;
    }
    else {
	fprintf(stderr,
		"\nERROR: (%d) NUMERATORS BUFFER OVERFULL IN storenumerator()\n\n",
		nnumerators);
    }
    return 0;
}

int print_ps_accid(FILE *fd)
{
    int i,n,a,b,d,s1,s2,c;
    unsigned m1,m2;
    char str[64];

    c=0;

    /*    for (i=0; i<nnumerators; i++)
	  fprintf(fd,"%d -> %d\n",i,numerators[i]); */

    if(denominator>256) {
	fprintf(stderr,"\nERROR: den>256 IN print_ps_accid()\n\n");
	denominator=256;
    }

    if(abcm2ps_den>2 && abcm2ps_den!=4) {
	fprintf(stderr,
		"\nWARNING: `abcm2ps:' [ %d ] in print_ps_accid().\n"
		"         User should provide PostScript accidentals for abcmp2s.\n\n",
		abcm2ps_den);
    }

    for (i=0; i<nnumerators; i++) {
	b = abcm2ps_den;
	d = denominator;
	n = numerators[i];
	s1 = n>0?1:(n?-1:0);
	/* a = (int) floor((n/(double)d) * b + .5); */
	a = (int) round((n/(double)d) * b);
	s2 = a>0?1:(a?-1:0);
	n = abs(n);
	a = abs(a);

	/* adjust fraction for quarter-tone and semitone */
	
	if(abcm2ps_reduction) {
	    if(a*2==b && (abcm2ps_den%2==0))
		a=1, b=2;
	    if(a==b)
		a=b=1;
	}

	m1 = microval(n,d);
	m2 = microval(a,b);

	if(m1==m2)
	    continue;

	if(a>b)
	    fprintf(stderr,
		    "\nERROR: FIXME: print_ps_accid()\n"
		    "       BAD QUANTIZATION!!!\n\n");

	c++;
	fprintf(fd,"%%%%postscript ");
	fprintf(fd,"/%s", s1>0?"sh":"ft");
	fprintf(fd,"%-5u ",m1);

	fprintf(fd,"/%s", s2>0?"sh":(s2?"ft":"nt"));

	/// if(s2 && b!=1)
	if(s2) /* not natural */
	    fprintf(fd,"%-5u ",m2);
	else
	    fprintf(fd,"0     ");
	    
	fprintf(fd,"load ! ");

 	/* output info */
	sprintf(str,"%c%d/%d",s1>0?'^':(s1?'_':'='),n,d);
	fprintf(fd,"%%  %10s",str);
	fprintf(fd,"  -->  %c",s2>0?'^':(s2?'_':'='));

	if(s2 && b!=1) /* not natural and not semitone */
	    fprintf(fd,"%d/%d",a,b);

	fprintf(fd,"\n");
    }
    return c;
}

unsigned microval(int n, int d)
{
    return (unsigned) ((256.0 * (n-1.0)) + (d-1.0));
}

double scl2cents(char *s) /* TODO: check-up */
{
    double n=0.0, d=1.0;
    int c;
    if((c=sscanf(s, "%lf/%lf\n",&n,&d))) {
	if(c==1 && strchr(s,'.')!=NULL) {
	    return n;
	}
    }
    return interv2cents(n/d);
}

int readscalafile(char *fname)
{
    FILE *fp;
    char s[BUFSIZ], t[BUFSIZ];
    double n,d,c;
    int i;
    int name, num;

    if((fp=fopen(fname,"rt"))==NULL) {
	fprintf(stderr,"\nERROR: CANNOT OPEN SCALA FILE `%s'\n\n",fname);
	exit (1);
    }

    name=num=0;
    scale[0]=0.0;
    i=1;

    /* TODO: capture name/description of the scale  */

    while (fgetline(s,BUFSIZ,fp)!=NULL) {
	if(sscanf(s,"!%s",t)) /* skip comment */
	    continue;
	else if(!name) { /* skip name */
	    name=1;
	    continue;
	}
	t[0]='\0';
	d=1.0;
	if(sscanf(s,"%s",t)<=0)
	    continue;
	if((c=sscanf(t,"%lf/%lf",&n,&d))>0) {
	    if(c==1 && strchr(t,'.')!=NULL)
		scale[i]=n;
	    else if (c==1 && num==0) { /* catch number of pitches */
		num=n;
		continue;
	    }
	    else if(d!=0.0)
		scale[i]=interv2cents(n/d);
	    else
		continue;
	    i++;
	    if(i>=MAX_SCALE_SIZE) {
		fprintf(stderr,"\nERROR: Scala file has too many pitches.\n\n");
		break;
	    }
	}
    }
    fclose(fp);

    /*    for (int w = 0; w<i; w++)
	  fprintf(stderr,"%d:\t%.5f\n",w,scale[w]); */

    if(num!=i-1) {
	fprintf(stderr,
		"\nWARNING: readscalafile(): Wrong number of pitches in \"%s\"\n",
		fname);
	fprintf(stderr,
		"         Number of pitches:  Declared: %d  Found: %d\n\n",num,i-1);
    }

    sclmodulo=scale[i-1];
    return i-1;
}


int output_psacc(FILE *fd)
{
    /* arrows */
    fprintf(fd,
	    "%%%%postscript /up-arrow{M 2 -7 RL -4 0 RL fill}!\n"
	    "%%%%postscript "
	    "/dn-arrow{gsave T 180 rotate 0 0 up-arrow grestore}!\n"
	    );
    /* 1/8-tone => natural + arrow */
    fprintf(fd,
	    "%%%%postscript "
	    "/sh3{2 copy nt0 exch 2 sub exch 14.5 add up-arrow}!\n"
	    "%%%%postscript "
	    "/ft3{2 copy nt0 exch 1.3 add exch 14.5 sub dn-arrow}!\n"
	    );
    /* 2/8-tone => use abcm2ps builtin quartertone accidentals */
    fprintf(fd,
	    "%%%%postscript /sh259/sh1 load !\n"
	    "%%%%postscript /ft259/ft1 load !\n"
	    );
    /* 4/8-tone => conventional sharp/flat */
    fprintf(fd,
	    "%%%%postscript /sh771/sh0 load !\n"
	    "%%%%postscript /ft771/ft0 load !\n"
	    );

    /*
     * For 3/8-tone there are two versions:
     */
    
    /*
     * (1) Version based on a conventional flat or sharp
     * 
     * 3/8-tone => sharp/flat + arrow
     * 
     * Easy to read as it avoids non-conventional accidentals
     *
     */

    fprintf(fd,
	    "%%%%postscript /sh515{2 copy sh0 exch 1.2 sub exch 15 sub dn-arrow}!\n"
	    "%%%%postscript /ft515{2 copy ft0 exch 1.8 sub exch 14 add up-arrow}!\n"
	    );

    /* 
     * (2) Version based on the direction of the accidental (up/down)
     *
     * 3/8-tone => quartertone + arrow
     *
     * Consistent with the direction of the alteration as it avoids
     * flats with arrow up or sharps with arrow down, but not as easy
     * to read as it is based on non-conventional accidentals
     */
    
    /*
      fprintf(fd,
	    "%%%%postscript "
	    "/sh515{2 copy sh259 exch 0 sub exch 14.5 add up-arrow}!\n"
	    "%%%%postscript "
	    "/ft515{2 copy ft259 exch 1.8 add exch 14.5 sub 2 copy\n"
	    "%%%%postscript "
	    "\tdn-arrow 2 add M 0 9 RL .8 setlinewidth stroke dlw}!\n"
	    );
    */

    /* (5/8-tone up to 8/8-tone accidentals are not generated) */

    return 1;
}

int output_timidity() /* Output 128 frequencies in miliHertz */
{
    int i;
    FILE *fd;
    fd=fopen(timidity_file,"wt");
    if(!fd) {
	fprintf(stderr,
		"\nERROR: CANNOT OPEN FILE %s FOR WRITTING (`timidity:')\n\n",
		timidity_file);
	return -1;
    }
    if(sname[0]!='\0')
	fprintf(fd,"# %s\n",sname);
    fprintf(fd,
	    "# Timidity++ tuning file generated by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n");
    for(i=0;i<128;i++) {
	fprintf(fd,"%.0lf\n",pitch2freq(i-centro)*1000.0); 
		/*fprintf(fd,"%.0lf\n",pitch2freq(i)*1000.0); */
    }

    fclose(fd);
    fprintf(stderr,
	    "TIMIDITY++ FREQUENCY FILE GENERATED: %s\n",
	    timidity_file);
    return 0;
}

int output_snt(FILE *fp) /* Output tuning file for abcMIDI */
{
    int i, openclosefile;

    if(fp==NULL) {
	openclosefile=1;
	fp=fopen(snt_file,"wt");
	if(fp==NULL) {
	    fprintf(stderr,
		    "\nERROR: CANNOT OPEN FILE %s FOR WRITTING (`abc2midi:')\n\n",
		    snt_file);
	    return -1;
	}
    }
    else
	openclosefile=0;
    if(sname[0]!='\0')
	fprintf(fp,"%% %s\n",sname);
    fprintf(fp,
	    "%% abc2midi snt tuning file generated by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n");
    for(i=0;i<128;i++)
	fprintf(fp,"%%%%MIDI snt %d %lf\n",
		/*i,freq2midi(pitch2freq(i),a_freq));*/
		i,freq2midi(pitch2freq(i-centro),a_freq));
    /* i,freq2midi(pitch2freq(i-centro),a_freq)); */
    /* TODO: check use of a_freq above */

    if(openclosefile)
	fclose(fp);
    fprintf(stderr,
	    "ABC2MIDI SNT FREQUENCY FILE GENERATED: %s\n",
	    snt_file);
    return 0;
}

int output_scala()
{
    int i;
    FILE *fd;
    fd=fopen(scala_file,"wt");
    if(!fd) {
	fprintf(stderr,
		"\nERROR: CANNOT OPEN FILE %s FOR WRITTING (`scala:')\n\n",
		scala_file);
	return -1;
    }
    fprintf(fd,
	    "! %s\n"
	    "! Scala tuning file generated by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n%s\n",
	    scala_file,scala_file);
    fprintf(fd," %d\n!\n",modulo);

#define FORMAT_STRING " %19.14lf\n"

    for(i=1;i<modulo;i++) {
	fprintf(fd,FORMAT_STRING,scale[i]);
    }
    fprintf(fd,FORMAT_STRING,sclmodulo); /*TODO: check*/
    fprintf(fd,"\n");
    fclose(fd);
    fprintf(stderr,
	    "SCALA FILE GENERATED: %s\n",
	    scala_file);
    return 0;

}

int output_scalakbm()
{

    /* TODO: support diatonic mapping */

    int i;
    FILE *fd;
    fd=fopen(scalakbm_file,"wt");
    if(!fd) {
	fprintf(stderr,
		"\nERROR: CANNOT OPEN FILE %s FOR WRITTING (`scalakbm:')\n\n",
		scalakbm_file);
	return -1;
    }
    fprintf(fd,
	    "! %s\n"
	    "! Scala keyboard mapping file generated by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n!\n",
	    scalakbm_file);

    /*
! Template for a keyboard mapping
!
! Size of map (greater than or equal to the number of notes in the scale 
! to be mapped). The pattern repeats every so many keys:
12
! First MIDI note number to retune:
0
! Last MIDI note number to retune:
127
    */

    fprintf(fd,"! Size of map:\n");
    fprintf(fd," %d\n!\n",modulo);
    fprintf(fd,"! First and last MIDI pitches to retune:\n");
    fprintf(fd," %d\n %d\n!\n",0,127);

    /*
! Middle note where scale degree 0 is mapped to:
60
! Reference note for which frequency is given:
69
! Frequency to tune the above note to (floating point e.g. 440.0):
440.0
! Scale degree to consider as formal octave (determines difference in pitch 
! between adjacent mapping patterns):
12
    */

    fprintf(fd,"! MIDI pitch for scale degree 0:\n");
    fprintf(fd," %d\n!\n",centro);
    fprintf(fd,"! Reference pitch:\n");
    fprintf(fd," %d\n!\n",centro);
    fprintf(fd,"! Reference frequency:\n");
    fprintf(fd," %lf\n!\n",basefreq);
    fprintf(fd,"! Scale degree to consider as formal octave:\n");
    fprintf(fd," %d\n!\n",modulo);

    /*
! Mapping.
! The numbers represent scale degrees mapped to keys. The first degree is for
! the given middle note, the next for subsequent higher keys.
! For an unmapped key, put in an "x". At the end, unmapped keys may be left out.
0
1
2
3
4
5
6
7
8
9
10
11
    */

    fprintf(fd,"! Mapping:\n");
    for(i=0;i<modulo;i++)
	    fprintf(fd," %2d\n",i);
    fprintf(fd,"\n");
    fclose(fd);
    fprintf(stderr,
	    "SCALA MAPPING FILE GENERATED: %s\n",
	    scalakbm_file);
    return 0;

}

static char *fgetline(char s[],int lim,FILE *ap_arquivo)
{
    int c,i,c2;

    s[0]=c='\0';

    for(i=0; i<lim-1
	    && (c=fgetc(ap_arquivo))!=EOF
	    && (!iseol(c));i++)
	s[i]=c;
    if(iseol(c)) { // TODO: check buffer full ??
	    s[i]=c;
	    i++;
	    if(c==13 && i<lim-1) { // treat EOL format for DOS
		if((c2=fgetc(ap_arquivo))!=EOF && c2==10) {
		    s[i++]=c2;
		    c=c2;
		}
		else
		    ungetc(c2,ap_arquivo);
	    }
    }

    s[i]='\0';

    if(c==EOF && strlen(s)==0) 
	return NULL;

    return s;
}

static int iseol(char c)
{
    switch(c) {
    case 10:
    case 13:
	return 1;
    default:
	return 0;
    }
}

int espacos(char *s)
{
    /* #include <ctype.h> */
    /* int isspace(int c); // function from ctype.h */
    int i=0;
    while(s[i]!='\0')
	if(!isspace(s[i++]))
	    return 0;
    return 1;
}

int preprocess(FILE *fin, FILE *fout, int outfor)
{
    int i, n, dlen, llen, rlen, slen, nlen;
    char s[BUFSIZ], out[BUFSIZ], name[BUFSIZ], subst[BUFSIZ];
    char *p=NULL, *p2=NULL, *p3=NULL;

    int intuneheader; /* inside a ABC tune header? (bool) */
    int modbak;
    
    fprintf(fout,"#\n# File preprocessed by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n#\n");

    if(procsag && outfor==FOR_PS && embed_sagittal_fmt) {
	fprintf(stderr,"EMBEDDING SAGITTAL TYPE 3 FONT\n");
	output_sagittal_fmt(fout); /* embed sagittal type 3 font */
    }

    llen = strlen(leftd);
    rlen = strlen(rightd);
    dlen=llen+rlen;
    intuneheader=0;
    modbak=modulo;
    while(fgetline(s,BUFSIZ,fin)!=NULL) { /* FOR EACH INPUT LINE */

	/* discard EOL characters */

	p=strchr(s,13); if(p!=0) *p = '\0'; /* descarta eol */
	p=strchr(s,10); if(p!=0) *p = '\0'; /* descarta eol */

	/* remove blank spaces to out[] */
	p=s;
	while(isspace(*p) && *p!='\0')
	    p++;
	strncpy(out,p,BUFSIZ-1);

	// TODO: command to insert tuning info (%%vskip / %%postscript textblock)

	// TODO: check!

	if(out[0]=='%' && out[1]=='%' && sscanf(out,"%%%%microabc: %s",name)==1) {
	    p = out+12;
	    ///	    fprintf(stderr,"%% READ :%s:\n",p);
	    //initialize();
	    /* moved to get_option() on lineartemp
	    generator = interv2cents(3.0/2.0);
	    force_modulo=0;
	    sclmodulo= 1200.0;
	    modulo=0;
	    equalt=lineart=0;
	    */
	    i = get_option(p);
	    //	    fprintf(stderr,"\n\n   READING OPTIONS! %s i=%d \a\n\n",s+12,i);
	    prepare();

	    if(i==0)
		get_proc_command(p,fin,fout); /* for `include:' */

	    //
	    // testing option to change the tuning on the fly
	    //
	    // TODO : treat blank spaces and ends of line for get_option
	    // (for some reason, EOL is omitted when printing s[]
	    // So I decided put the `continue' statement.
	    //
	    continue;
	}

	if(sscanf(s,"K:%s",out)==1 && intuneheader) {
	    intuneheader = 0;

	    /* TODO: Include information on tuning in ABC header */

	    ///// TODO: decide if maintain use of `%%MIDI snt'
	    //// Would be good because #include is not supported
	    if( 0 ) {
		if(modulo!=12 || sclmodulo!=1200.0) {
		    if (equalt) {
			/* temporarily change scale */
			modbak=modulo;
			fprintf(fout,
				"%% Scale using "
				"equivalence inverval of %.3lf cents "
				"and generator of %.3lf cents\n",
				sclmodulo,scale[set_12out_et(0)]);
			/* SET BASE TUNING FOR abc NOMINALS */
			output_snt(fout);
			/* restaure original scale */
			equaltempc(modbak,(sclmodulo),scale);
			modulo=modbak;
		    }
		    else {
			/* SET BASE TUNING FOR abc NOMINALS */
			output_snt(fout);
		    }
		}   
	    }
	    ///////
	    
	}
	if(sscanf(s,"X:%d",&n)==1) {
	    intuneheader = 1;
	}

	if(procmac) {
	    
	    /* PREPROCESS USER MACROS (BEGIN) */
	    p=s;
	    slen=strlen(s);
	    out[0]='\0';
	    while(p<(s+slen)) { /* FOR EACH MACRO FOUND */
		
		p2=p3=NULL;
		
		if((n=(strlen(p))<=dlen)) {/* skip too short lines */
		    strncat(out,p,BUFSIZ-1-strlen(out));
		    break;
		}
		if(strstr(p,leftd)==NULL
		   || strstr(p,rightd)==NULL) {/* skip lines without
						  delimiters */
		    strncat(out,p,BUFSIZ-1-strlen(out));
		    break;
		}
		
		/* find left delimiter */
		p2=strstr(p,leftd);
		
		/* output leading text */
		n=strlen(out);
		while(p<p2) {
		    /*fputc(*p,fout);*/
		    out[n++]=*p;
		    p++;
		}
		out[n]='\0';
		
		/* find right delimiter */
		p3=strstr(p2+llen,rightd);
		/* if not found right delimiter */
		if(p3==NULL) {
		    strncpy(name,p2+llen,BUFSIZ-1);
		}
		else {
		    /* mount hypothetical macro name */
		    n = p3-p2-llen; // TODO: check this with long delimiters
		    if (n>BUFSIZ-1)
			n=BUFSIZ-1;
		    strncpy(name, p2+llen, n);
		    name[n]='\0';
		}

		nlen=strlen(name);

		if(usingsagittalalias)
		    normalizesagittal(name);
		
		/* look for macro definition */
		subst[0]='\0';
		for(i=0; i<nmacrodefs; i++)
		    if(strcmp(name,macrodef[i][0])==0) {
			strncpy(subst,macrodef[i][1],BUFSIZ-1);
			break;
		    }

		if(subst[0]!='\0') { /* found */
		    strncat(out,subst,BUFSIZ-1-strlen(out));
		    p+=(llen+rlen+nlen);
		}
		else { /* output non-macro */
		    n=strlen(out);
		    while(p<p2+llen) {
			out[n++]=*p;
			p++;
		    }
		    out[n]='\0';
		}
		
	    } /* FOR EACH MACRO */
	    /* PREPROCESS USER MACROS (END) */
	    
	}
	else
	    strncpy(out,s,BUFSIZ-1);

	if(procsag) {
	    /* CONVERT SAGITTAL MACROS/ACCIDENTALS */
	    /* They're processed AFTER common macros so that common macros
	       can use sagittal accidentals to be interpreted */

	    sagittalprocmac(out,BUFSIZ,outfor);

	    if(outfor==FOR_MIDI)
		sagittalprocabc(out,BUFSIZ);
	    /* TODO: treat octave number for Sagittal pitches in the
	       form ^/196F for PS */

	}

	fprintf(fout,"%s\n",out); /* PRINT OUT PROCESSED LINE (with \n) */
	
    } /* FOR EACH LINE */
    
    return 1;
}

int prepare_sagittal()
{
    int i;
    double scale_bak[MAX_SCALE_SIZE];
    double sclmodulo_bak;
    double modulo_bak;
    double generator_bak;
    double equalt_bak;
    double lineart_bak;

    if(sagittal2scale) { /* approximate sagittal -> scale */

	/* backup data temporarily for independent sagittal settings */
	for (i=0; i<MAX_SCALE_SIZE; i++)
	    scale_bak[i] = scale[i];
	sclmodulo_bak = sclmodulo;
	modulo_bak = modulo;
	generator_bak = generator;
	equalt_bak = equalt;
	lineart_bak = lineart;

	/* set tempering parameters for sagittal */
	sclmodulo = sag_sclmodulo;
	modulo = sag_modulo;
	generator = sag_generator;
	equalt = sag_equalt;
	lineart = sag_lineart;

	/* generate lineartemp or equaltemp for sagittal */
	if(sagittal2scale > 0)
	    equaltempc(modulo,(sclmodulo),scale);
	else
	    lineartempc(modulo,(sclmodulo),
			(generator),
			(lineart<0 ? 0 : lineart),
			scale);

	set_temper();  /* set temper for sagittal */

	/* restore original scale settings */
	for (i=0; i<MAX_SCALE_SIZE; i++)
	    scale[i] = scale_bak[i];
	sclmodulo = sclmodulo_bak;
	modulo = modulo_bak;
	generator = generator_bak;
	equalt = equalt_bak;
	lineart = lineart_bak;
    }
    else
	set_temper();  /* just set temper for sagittal */

    return 0;
}

int set_temper()
{
    int i;
    double f;

    /* set tempering parameters for sagittal */
    if(equalt) {
	for(i=0; i<NPRIMES; i++) {
	    /* temper primes (quantize in equal temperament) */
	    temper[i] =
		scale[mod((int) (round(f_mod(interv2cents(primes[i]),1200.0)
				      /scale[1])),modulo)]
		    - f_mod (interv2cents(primes[i]),1200.0);
	}
    }
    else if(lineart) { /* lineartemp */
	/* temper octave */
	/* assumes that sclmodulo is an octave */
	temper[0] = sclmodulo-1200.0;
	
	/* temper fifth */
	/* int n;
	 n = bestfifthindex(); 
	 f = (scale[n]+sclmodulo); */ /* tempered 3/1 */
	/* assumes that generator is a fifth */
	f = generator + sclmodulo; /* tempered 3/1 */
	temper[1] = f - interv2cents(3.0/1.0);
	
	for(i=2; i<NPRIMES; i++)
	    temper[i]=0.0;
    }
    return 0;
}

int sagittalprocabc(char s[], int lim) /* replace sagittal ABC microtones in s[] */
{
    char dest[64], out[BUFSIZ];
    char *p, *p2;
    char nominal, updn;
    int code, n, slen;
                
    /* TODO: Make this an user option */
    int pyth = 1; /* Option to use/avoid convert pythagorean nominals */

    /*    if(espacos(s))
	  return 1; */
                
    p=s;    
    slen=strlen(s);
    out[0]=out[1]='\0';
    while(p<(s+slen)) { /* FOR EACH SAGITTAL ACCIDENTAL FOUND */
        
	//	printf("\n1) out=%s:n=%d",out,n);
        p2=NULL;

        p2 = strchr(p,'/');
        if(p2!=NULL && p2>p) {
	    if( ( *(p2-1)=='^' ||
		     *(p2-1)=='=' ||
		     *(p2-1)=='_' )) {
		p2--;
	    }
	    else {

		strncat(out,p,p2-p+1); /// ???
		p=p2+1;

		continue; /// TODO: check this!!
	    }
	}
	if(p2==NULL) { /* NO MORE: COPY REMAINING TEXT */
	    strncat(out,p,BUFSIZ-1-strlen(out));
	    break;
	}
	/* output leading text */
        n=strlen(out);
	//	printf("\n2) out=%s:n=%d",out,n);
        while(p<p2) {
            out[n++]=*p;
            p++;
        }
        out[n]='\0';
	///	printf("\n3) out=%s:n=%d",out,n);

        dest[0]=nominal=updn=code=0;
        if(sscanf(p2+2,"%d%c%c",&code,&nominal,&updn)>=2
           && sagcode2abc(dest,code,nominal,denominator,pyth,temper) ) {
            n=strlen(dest);
            if(updn==',' || updn=='\'') {
                dest[n++]=updn;
                dest[n++]='\0';
                p2 = strchr(p2+2,updn);
            }
            else
                p2 = strchr(p2+2,nominal);
            p2++;
            fixabc(dest);
            strncat(out,dest,BUFSIZ-1-strlen(out));

        }
        else {
            /* copy the two chars ^/ =/ or _/ */
            n=strlen(out);
            out[n++]=*(p2);
            out[n++]=*(p2+1);
            out[n]='\0';
            p2+=2;
        }
        p=p2;
    }

    strncpy(s,out,lim-strlen(s));
    s[lim-1]='\0';

    return 1;
}

/* replace sagittal pitch macros in s[] */
int sagittalprocmac(char s[], int lim, int outfor)
{
    char  out[BUFSIZ], subst[BUFSIZ], name[BUFSIZ], sub[BUFSIZ];
    char temp[BUFSIZ];
    char *p, *p2, *p3;
    int  n, slen, llen, rlen, dlen, pyth;
    static double maxerr = 0.0;
    double err;
    char nomin[12];
    sagpitch pitch;

    double xx, fabc, f0, nearest, ftemp;
    int mm, nn, nearestpitch, ptemp;

    strcpy(nomin,"ABCDEFG");

    /*

    -- fixsagittal() (like fixabc())
    -- better support for octaves as numbers (fixnumericalupdn())
    -- valid nominals (scale not YWUNMX -- option to make these valid )
    -- should the scale be defined only by nominals?
    -- equal temperaments need rounding up accidentals
    -- user should be able to define pitches in Sagittal format
    (`aliassag:', `replacesag:')
    -- `defacc:' with meaning (interval) to define accidentals?
    -- to generate scores, a sagittal pitch can be replaced for
    something very different: 7-ET can be written as 7 nominals
    without accidentals, and then printed using 56-ET pitches.
     
    ./|\..  ./|\.  ./|\    ./|\'   ./|\''  /|\..  /|\.  

    /|\     /|\'   /|\''  '/|\..  '/|\.  '/|\   '/|\'  '/|\''

    C#'/|
    '/|#C     <====
    C#'/|5        |
    '/|#C6    <====
    C#'/| 5       |===== reverse core mixed accidental
    '/|#C 6   <====
    C#'/| -5      |
    '/|#C:6   <====

    */


    llen = strlen(leftd);
    rlen = strlen(rightd);
    dlen=llen+rlen;
    
    p=s;
    slen=strlen(s);
    out[0]='\0';
    while(p<(s+slen)) { /* FOR EACH MACRO FOUND */

	p2=p3=NULL;

	if((n=(strlen(p))<=dlen)) {/* skip too short lines */
	    strncat(out,p,BUFSIZ-1-strlen(out));
	    break;
	}
	if(strstr(p,leftd)==NULL
	   || strstr(p,rightd)==NULL) {/* skip lines without
					  delimiters */
	    strncat(out,p,BUFSIZ-1-strlen(out));
	    break;
	}

	/* find left delimiter */
	p2=strstr(p,leftd);
	/* output leading text */
	n=strlen(out);
	while(p<p2) {
	    /*fputc(*p,fout);*/
	    out[n++]=*p;
	    p++;
	}
	out[n]='\0';

	/* find right delimiter */
	p3=strstr(p2+llen,rightd);
	/* mount hypothetical macro name */
	n = p3-p2-llen;
	n = n>(BUFSIZ-1)?(BUFSIZ-1):n;
	strncpy(name, p2+llen, n);
	name[n]='\0';
	/* look for macro definition */
	subst[0]='\0';


	if(setsagpitch(&pitch,name,nomin,(unsigned char) 0377)) {
	    strncpy(subst,pitch.name,BUFSIZ-1);

	    /*
	    if(inspect) {
		fprintf(stderr,"\nPARSED SAGITTAL PITCH MACRO: %s\n",name);
		fprintf(stderr,"pitch.name=           %s\n",pitch.name);
		fprintf(stderr,"pitch.nominal= %c\n",pitch.nominal);
		fprintf(stderr,"pitch.octave= %d\n",pitch.octave);
		fprintf(stderr,"pitch.accidental.code= %d\n",pitch.accidental.code);
		fprintf(stderr,"code %d means %.lf/%.lf = %.3lf cents\n",
			pitch.accidental.code,
			sagittal[pitch.accidental.code].num,
			sagittal[pitch.accidental.code].den,
			sagittal[pitch.accidental.code].cents);
		fprintf(stderr,"pitch.accidental.left= %s\n",pitch.accidental.left);
		fprintf(stderr,"pitch.accidental.right= %s\n",pitch.accidental.right);
	    }
	    */

	    if(outfor==FOR_PS) {
		/* Convert some mixed symbols (Only for Sagittal version 2) */
		/* TODO: write a function for this, using MACROS for the constants */
		if (avoidmixedsag) { /* There are redundances in CharMap2 */
		    switch(pitch.accidental.code) {
		    case  33: /* double flat */
			if(strcmp(pitch.accidental.name,"bb")==0)
			    pitch.accidental.code=40;
			break;
		    case  35: /* flat */
			if(strcmp(pitch.accidental.name,"b")==0)
			    pitch.accidental.code=92;
			break;
		    case 253: /* sharp */
			if(strcmp(pitch.accidental.name,"#")==0)
			    pitch.accidental.code=196;
			break;
		    case 255: /* double sharp */
			if(strcmp(pitch.accidental.name,"x")==0)
			pitch.accidental.code=248;
			break;
		    default:
			;
		    }
		}
		if(pitch.accidental.code==SAGSPACE
		   || pitch.accidental.visible==0)
		    sprintf(subst,"%c",pitch.nominal); /* print only nominal */
		else
		    sprintf(subst,"%c/%d%c",
			    pitch.accidental.code>=SAGNATURAL ? '^' : '_',
			    pitch.accidental.code,
			    pitch.nominal);
	    }
	    else { /* FOR_MIDI */

		if(pitch.accidental.code==0) /* force natural accidental */
		    pitch.accidental.code=SAGNATURAL;

		/* Convert some mixed symbols (Only for Sagittal version 2) */
		/* TODO: write a function for this, using MACROS for the constants */
		switch(pitch.accidental.code) {
		case  33: /* double flat */
		    pitch.accidental.code=40;
		    break;
		case  34: /* Couper's sesquiflat */
		    pitch.accidental.code=68;
		    break;
		case  35: /* flat */
		    pitch.accidental.code=92;
		    break;
		case  36: /* Couper's semiflat */
		     pitch.accidental.code=120;
		    break;
		case 37: /* Wilson's minus*/
		    pitch.accidental.code=134;
		    break;
		case 251: /* Winson's plus */
		    pitch.accidental.code=154;
		    break;
		case 252: /* Tartini's semisharp */
		    pitch.accidental.code=168;
		    break;
		case 253: /* sharp */
		    pitch.accidental.code=196;
		    break;
		case 254: /* Tartini's sesquisharp */
		    pitch.accidental.code=220;
		    break;
		case 255: /* double sharp */
		    pitch.accidental.code=248;
		    break;
		default:
		    ;
		}

		/* Some non-Sagittal symbols (wilson / accents) */
		/* TODO: + and - used to make compound nominals in
		   combination with the letters A thru G, in the
		   chain-of-generators (CoG) notations for linear
		   temperaments.*/
		if( (pitch.accidental.code >= 38 && pitch.accidental.code <=39)
		    || (pitch.accidental.code >= 249 && pitch.accidental.code <=250))
		    fprintf(stderr,
			    "WARNING: (Non-)Sagittal accidental (code %d) "
			    "indefined for MIDI: %s\n",
			    pitch.accidental.code,pitch.name);

		pyth=1; /* boolean for adjusting nominals or not */
		if(sagcode2abc(subst,
			       pitch.accidental.code,
			       pitch.nominal,
			       denominator,
			       pyth,
			       temper)) {

		}
	    }
	    n=pitch.octave;
	    while(n!=0)
		if(pitch.octave<0)
		    strncat(subst,",",BUFSIZ-1-strlen(subst)), n++;
		else
		    strncat(subst,"'",BUFSIZ-1-strlen(subst)), n--;
	    fixabc(subst);
	}

	if(subst[0]!='\0') { /* found */
	    if(outfor==FOR_MIDI) {

		sagcode2abc(sub,
			    ' ',
			    sagittal_base[0],
			    denominator,
			    1,
			    temper);

		/////////////
		/* TODO: clean up the code below 
		   (there are redundances) */
		if(abcm2ps_den)
		    freq2abc(temp,
			     abc2freq(subst)*(440.0/a_freq)
			     /(abc2freq(sub)/abc2freq(sagittal_base)
			       /(sagittal_basefreq/abc2freq(sagittal_base))));
		else
		    freq2abc(temp,
			     abc2freq(subst)*(a_freq/440.0)
			     /(abc2freq(sub)/abc2freq(sagittal_base)
			       /(sagittal_basefreq/abc2freq(sagittal_base))));
		
		if(scl>=0 && sagittal2scale) { /* approximate to
						 nearest pitch in the
						 scale */
		    /* TODO: turn this a function nearestpitch() */
		    fabc = abc2freq(temp);
		    f0 = pitch2freq(0);
		    if(inspect)
			fprintf(stderr,"sub=\"%s\" subst=\"%s\" temp=\"%s\" "
				"fabc=%.lfHz f0=%.lfHz\n",
				sub, subst, temp,fabc,f0);
		    xx = interv2cents(fabc/f0);
		    err = 0.0;
		    nn=0;
		    while(xx>=sclmodulo)
			xx-=sclmodulo, nn++;
		    while(xx<0)
			xx+=sclmodulo, nn--;

		    ftemp = nearest = 1.0;
		    ptemp  = nearestpitch = 0;

		    for(mm = -modulo*2; mm<=modulo*2; mm++) {
			ptemp = mm+nn*modulo;
			ftemp = pitch2freq(ptemp);
			if(fabs(interv2cents(ftemp/fabc))
			   <= fabs(interv2cents(nearest/fabc))) {
			    nearest = ftemp;
			    nearestpitch = ptemp;
			}
		    }
		    err = interv2cents(nearest/fabc);
		    if(fabs(err) > fabs(maxerr))
			maxerr = err;
		    if(inspect) {
			fprintf(stderr,
				"\nApproximating fabc:  %5.2lf Hz"
				" to nearest:  %5.2lf Hz\n"
				"Difference:    %7.2lf cents\n",
				fabc, nearest,
				err);
			fprintf(stderr,
				"Highest diff.: %7.2lf cents\n",
				maxerr);
		    }

		    if(outputmode==LITERAL) {
			alias(temp,nearestpitch,replace);
			if(inspect)
			    fprintf(stderr,"LITERAL\n");
		    }
		    else if(outputmode==MICROTONAL) {
			if(inspect)
			    fprintf(stderr,"MICROTONAL\n");
			////alias(temp,nearestpitch,scl);
			freq2abc(temp,nearest);
		    }
		    else if (outputmode==DIATONIC || outputmode==CHROMATIC) {
			if(inspect)
			    fprintf(stderr,"%s\n",
				    outputmode==DIATONIC?"DIATONIC":"CHROMATIC");
			//abc(temp,nearestpitch+centro, ///////
			//    outputmode);
			strcpy(temp,"\0");
			abc(temp,nearestpitch+centro,outputmode);
		    }
		    else {
			if(inspect)
			    fprintf(stderr,"UNKNOWN\n");
			/* ??? */;
		    }
		}
		
		strcpy(subst,temp);

	    }
	    
	    if(inspect)
		fprintf(stderr,"%15s   =>   %15s    ( %5.2lf Hz )\n",
			name,subst,abc2freq(subst));
	    strncat(out,subst,BUFSIZ-1-strlen(out));
	    p+=(llen+rlen+strlen(name));
	}
	else { /* output non-macro */
	    n=strlen(out);
	    while(p<p2+llen) {
		out[n++]=*p;
		p++;
	    }
	    out[n]='\0';
	}
    } /* FOR EACH MACRO */

    strncpy(s,out,lim-strlen(s));
    s[lim-1]='\0';

    return 1;
}

int normalizesagittal(char *s)
{
    sagpitch pitch;
    char name[MAXSAGNAME];
    char nomin[12];
    const char sagup[2]={SAGOCTAVEUP,'\0'};
    const char sagdn[2]={SAGOCTAVEDN,'\0'};
    int i, octisnum;
    
    /* TODO: normalize octave as number (difficult to adjust octave numbers) */

    strcpy(nomin,"ABCDEFG"); /* TODO: synchronize this with
				sagittalprocmac or make it global */
    strncpy(name,s,MAXSAGNAME-1);
    cleansagpitch(&pitch);

    if(setsagpitch(&pitch,name,nomin,(unsigned char) 0377)) {
	i=0;
	/* Build string following this order: */
	/*   invisible */
	if(pitch.accidental.visible==0)
	    name[i++]=DONTSHOW;
	/*   nominal */
	name[i++]=pitch.nominal;
	name[i]='\0';
	/*   left accents */
	strncat(name,pitch.accidental.left,MAXSAGNAME-1-strlen(name));


	/*   core accidental */
	strncat(name,
		/* sagittal[pitch.accidental.code].longpure, */
		pitch.accidental.name,
		MAXSAGNAME-1-strlen(name));

	/*   right accents */
	strncat(name,pitch.accidental.right,MAXSAGNAME-1-strlen(name));

	/*   octave modifiers (`` ,,) */
	i = pitch.octave;
	while(i!=0) {
	    if(i>0) {
		strncat(name,sagup,MAXSAGNAME-1-strlen(name));
		i--;
	    }
	    else {
		strncat(name,sagdn,MAXSAGNAME-1-strlen(name));
		i++;
	    }
	}
    }
    else
	return 0; /* failed */
    strncpy(s,name,MAXSAGNAME-1-strlen(s));
    return 1; /* succeded */
}

int includefile(FILE *fout,char name[])
{
    FILE *fin;
    int c=0;
    if((fin=fopen(name,"rt"))==NULL) {
	fprintf(stderr,"\nERROR: includefile() could not open `%s'\n\n",name);
	return 0;
    }
    fprintf(fout,"%% (microabc) File `%s' included below:\n",name);
    while((c=fgetc(fin))!=EOF)
	fputc(c,fout);
    fprintf(fout,"%% (microabc) File `%s' included above.\n",name);
    fclose(fin);
    return 1;
}

int set_meantone()
{
    modulo=12;
    sclmodulo=1200.0;
    scl=10000;
    /* a meantone scale */
    /* TODO: use function lineartempc() */
    scale[0]= 0.0;
    scale[1]= 86.0;
    scale[2]= 196.0;
    scale[3]= 306.0;
    scale[4]= 392.0;
    scale[5]= 502.0;
    scale[6]= 588.0;
    scale[7]= 698.0;
    scale[8]= 784.0;
    scale[9]= 894.0;
    scale[10]= 1004.0;
    scale[11]= 1090.0;
    return 1;
}

int set_pythagorean()
{
    modulo=12;
    sclmodulo=1200.0;
    scl=10000;
    /* a pythagorean scale */
    lineartempc(modulo,sclmodulo,interv2cents(3.0/2.0),3,scale);
    return 1;
}

int set_12out_et(int best) /* return value is the best fifth index */
{
    /* To compute nominals for Sagittal ET notation */
    /* filling all 12 pitches for output_snt(fout). */
    /* Assumes scale[] already has the correct data */


    /*
     *
     * TODO: SAVE OLD VALUE OF MODULO
     *
     */

    int i;
    double sc[12];

    /* find best fifth */
    if(best==0)
	best = bestfifthindex();

    /* just for curiosity */
    if((modulo % (best+1)) == 0) {
	fprintf(stderr,
		"\a\nBest fifth (%d steps = %lf cents) "
		"is a divisor of octave in %d-EDO.\n\n",
		best, scale[best], modulo);
    }
    else
	fprintf(stderr,
		"\nBest fifth is %d steps = %lf cents in %d-EDO.\n\n",
		best, scale[best], modulo);
    fprintf(stderr,"temper: %lf %lf\n",temper[0],temper[1]);

    /* make a chain of best fifths */
    sc[0] = scale[0];   		/* C */
    sc[7] = scale[best];		/* G */
    sc[2] = scale[mod(best*2,modulo)]; /* D */

    sc[9] = scale[mod(best*3,modulo)]; /* A */
    sc[4] = scale[mod(best*4,modulo)]; /* E */
    sc[11]= scale[mod(best*5,modulo)]; /* B */

    sc[6] = scale[mod(best*6,modulo)]; /* F# */
    sc[1] = scale[mod(best*7,modulo)]; /* C# */
    sc[8] = scale[mod(best*8,modulo)]; /* G# */

    sc[3] = scale[mod(best* -3,modulo)]; /* Eb */
    sc[10]= scale[mod(best* -2,modulo)]; /* Bb */
    sc[5] = scale[mod(best* -1,modulo)]; /* F */

    /* copy scale and replace modulo with 12 */
    for(i=0;i<12;i++) {
	scale[i]=sc[i];
	fprintf(stderr,"scale[%d]=%lf\n",i,scale[i]);
    }
    modulo=12;

    return best;
}

int bestfifthindex()
{
    int best, i;
    double fifth;

    fifth = (700.0+FIFDIF); /* 3/2 as cents -- FIFDIF from interval.c */

    /* assume scale was set to temperament */
    for(i=0;i<modulo;i++) {
	if(scale[i] > fifth)
	    break;
    }
    if(i>=modulo)
	i=modulo-1;
    /* assume fifth is not in the bounds of scale */
    if(fabs(fifth-scale[i]) < fabs(fifth-scale[i-1]))
	best=i;
    else
	best=i-1;

    return best;
}

int output_testfile(char *fname)
{
    FILE *fp;
    int i, j, m, skip, w, t, midi, outcont;
    char buf[BUFSIZ], buf2[BUFSIZ], *p;

    outcont=0;
    
    if((fp=fopen(fname,"wt"))==NULL) {
	fprintf(stderr,"ERROR: CANNOT OPEN TEST FILE %s FOR WRITTING\n\n",fname);
	return -1;
    }

    /* scape \ % and " for abc string */
    strcpy(buf,commandline);
    j=0; p=buf2;
    while(buf[j] != '\0') {
	if(buf[j] == '\\' || buf[j] == '"' || buf[j] == '%')
	    *p++ = '\\';
	*p++ = buf[j];
	j++;
	if(p-buf2 >= BUFSIZ)
		break;
    }
    *p='\0';

    fprintf(fp,
	    "%% Test file generated by "
	    "microabc (C) 2006-2008 Hudson Lacerda\n\n"
	    );

    if(procsag || replace_sagittal || usingsagittalalias) {
	fprintf(fp,"%%%%format sagittal.fmt\n");
	fprintf(fp,"%%%%format sagittal-mixed.fmt\n");
	fprintf(fp,"%%%%postscript "
		"currentdict/sagpure known true eq{sagpure}if\n"
		"\n%% If you want mixed Sagittal notation, \n"
		"%% let exactly two '%%%%' at the start of the following line:\n"
		"%%%%%%postscript sagmixed\n"
		"\n");
    }

    fprintf(fp,
	    // "%%%%annotationfont Courier-Bold 9\n"  /* font family */
	    //   "%%%%annotationfont Courier 11\n"  /* font family */
	    "%%%%scale 0.8\n" /* magnification */
	    "%%%%continueall true\n" /* auto formatting */
	    "%%%%maxshrink 0.4\n" /* auto formatting */
	    "%%%%measurenb 0\n"      /* measure numbers */
	    "%%%%staffsep 60pt\n" /* separate staves */
	    "\n"
	    );

    fprintf(fp,
	    "%%%%begintext fill\nCOMMAND LINE: %s\n%%%%endtext\n"
	    "%%%%text MICROABC INPUT FILE: %s\n"
	    "%%%%text MICROABC MODE: %s\n"
	    "%%%%text SCALA *.scl INPUT FILE: %s\n"
	    "%%%%text SCALA *.scl OUTPUT FILE: %s%s\n"
	    "%%%%text SCALA *.kbm OUTPUT FILE: %s\n"
	    "%%%%text TIMIDITY OUTPUT FILE: %s%s\n"
	    "%%%%text SCALE SIZE: %d\n"
	    "%%%%text RANGE: %d %d (%d)\n"
	    "%%%%text NUMBER OF DEFINITIONS: %d\n"
	    "\n",
	    buf2, infile,
	    outputmode==LITERAL?"literal":
	    (outputmode==CHROMATIC?"chromatic":
	     (outputmode==MICROTONAL?"microtonal":
	      (outputmode==DIATONIC?"diatonic":
	       "ERROR: unknown mode!!!"))),
	    scala_input_file,
	    scala_file,(outscala&&scl<0)?" [FAILED]":"",
	    scalakbm_file,
	    timidity_file,(timidity&&scl<0)?" [FAILED]":"",
	    modulo,
	    amin-centro,amax-centro,(1+amax-amin),
	    nmacrodefs
	    );

    /* ignore `dumcacc:' and sname[] because they are not ABC notes */
    if(sname[0]!='\0')
	skip=1;
    else
	skip=0;
    if(accid_dmacros)
	skip+=naccid + 5; /* There are 5 builtin accidentals (^^ ^ = _ __) */


    /* PRINT OUT NON-NOTE MACROS (sname[] and dumbacc:) */
    if(skip>0) {
	fprintf(fp,"\n");
	fprintf(fp,"%%%%begintext obeylines\n");
	fprintf(fp,"SPECIAL DEFINITIONS:\n");
	for (i=0; i<skip; i++) {
	    for(w=0; w<2; w++) {
		strcpy(buf,macrodef[i][w]);
		/* scape \ % and " for abc string */
		j=0; p=buf2;
		while(buf[j] != '\0') {
		    if(buf[j] == '\\' || buf[j] == '"' || buf[j] == '%')
			*p++ = '\\';
		    *p++ = buf[j];
		    j++;
		    if(p-buf2 >= BUFSIZ)
			break;
		}
		*p='\0';
		fprintf(fp,"      %s",buf2);
	    }
	    fprintf(fp,"\n");
	}
	fprintf(fp,"%%%%endtext\n");
    }

    /* PRINT OUT MUSICAL NOTES */
    fprintf(fp,
	    "X:1\n"
	    "T:Test file generated by microabc\n"
	    "C:microabc (C) 2006-2008\n"
	    "C:Hudson Lacerda\n"
	    "%%%%MIDI nobeataccents\n"
	    "L:1/4\nM:4/4\n"
	    "K:C\n"
	    );
    m=1; /* measure number */
    t=0; /* measure beat */
    fprintf(fp,"%% %d\n",m++);
    for (i=skip; i<nmacrodefs; i++) {
	strcpy(buf,macrodef[i][0]);

	/* scape \ ; and " for abc string */
	j=0; p=buf2;
	while(buf[j] != '\0') {
	    if(buf[j] == '\\' || buf[j] == '"' || buf[j] == ';')
		*p++ = '\\';
	    *p++ = buf[j];
	    j++;
	    if(p-buf2 >= BUFSIZ)
		break;
	}
	*p='\0';

	fprintf(fp,"\"^;^;^%s\"",buf2); /* macro name (3 lines) */

	strcpy(buf,macrodef[i][1]); /* check replacement */

	if(!inspect) {
	    if(buf[0]!='\0')
		midi=freq2midi(abc2freq(buf),440.0);
	    else
		midi=-1;
	}
	else
	    midi=0;

	if(midi < 0 || midi > 127 ) {
	    outcont++;
	    strcpy(buf,"\"_Out of;range!\"\tx"); /* warning on invisible rest */
	    fprintf(fp,"%s\n",buf);
	}
	else
	    fprintf(fp,"\t%s%s%s\n",leftd,macrodef[i][0],rightd);
	
	if(t%4==3 && i!=(nmacrodefs-1)) {
	    fprintf(fp,"|\n");
	    fprintf(fp,"%% %d\n",m++);
	}
	t++;
    }

    if(t==0)
	fprintf(fp,"z4\n");
    else if(t%4) {
	switch(4-(t%=4)) {
	case 1:
	    fprintf(fp,"z\n");
	    break;
	case 3:
	    fprintf(fp,"z\n");
	case 2:
	    fprintf(fp,"z2\n");
	}
    }

    fprintf(fp,"|]\n");

    if(outcont) {
	fprintf(fp,"\n%%%%text WARNING: testfile(): "
		"There are %d pitches out of MIDI range.\n\n",
		outcont);
	fprintf(stderr,"\nWARNING: testfile(): "
		"There are %d pitches out of MIDI range.\n\n",
		outcont);
    }
    
    fprintf(fp,
	    "\n\n%% End of test file.\n"
	    );
    fclose(fp);

    fprintf(stderr,"TEST FILE GENERATED: %s\n",fname);

    return 0;
}

int clean_string(char *s) /* remove leading and trailing spaces */
{
    char *p, t[BUFSIZ];
    int len;
    
    /* remove trailing spaces */
    while((len=strlen(s)) && isspace(s[len-1]))
	s[len-1]='\0';

    /* remove leading spaces */
    strncpy(t,s,BUFSIZ-1);
    p=t;
    while(isspace(*p) && *p!='\0')
	p++;
    strcpy(s,p);

    return 0;
}


