/* microabc.h -- this file is part of microabc (C) 2006-2008 Hudson Lacerda
 * Generates a file for the preprocessor abcpp, with MIDI mapping or
 * ABC microtonal notation, intended to notate arbitrary scales and
 * tunings.
 */

/*
 * Copyright (C) 2006-2008 Hudson Flvio Meneses Lacerda
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _MICROABC__H_
#define _MICROABC__H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h> /* link to it with the gcc option -lm */
#include "interval.h"
#include "sagittal.h"
#include "sagdata.h"

#define MAX_DEFINITIONS 4096	/* Max number of macro actual definitions */
#define MAX_ALIAS         10	/* Max number of input columns */
#define MAX_MACROS      1200	/* Max number of macros (scale degrees) */
#define MAX_SCALE_SIZE  1200	/* Max size for scale */
#define MAX_NUMERATORS   512	/* Max different accidentals for abcm2ps */
#define MAX_ACCID         64	/* Max accidental definitions */
#define MAX_NAME_LENGHT  255    /* Max length of a name (without count '\0') */
#define MAX_BUFLINES     128    /* Max size of textbuf[][] */
#define MAX_BUFLINELEN   128    /* Max length of a line in textbuf[][] */

#define OCTSEP '.'  /* Default octave number separator */

/* output format for internal preprocessor */
#define FOR_MIDI 0
#define FOR_PS   1

/* math.h */
#ifdef CYGWIN /* long-double functions disabled for CYGWIN */
#ifndef powl
#define powl(x,y) pow((double)(x),(double)(y))
#endif
#ifndef log2l
#define log2l(x) (log((double)(x)) / log(2.0))
#endif
#ifndef exp2l
#define exp2l(x) pow(2.0,(double)(x))
#endif
#ifndef trunc
#define trunc(x) (  (double)   ( (long int)(x) )  )
#endif
#ifndef round
#define round(x) ( (double) ( (long int)((x) +  ((x)>0?.5:(-.5)))))
#endif
#else /* not CYGWIN */
long double powl(long double x, long double y);
long double log2l(long double x);
long double exp2l(long double x);
double round(double);
double trunc(double);
#endif

/* extern (interval.[ch]) */
double cents2interv(double cents);
double freq2midi(double freq, double a_freq);
double interv2cents(double interv);
double midi2freq(double midi, double a_freq);
int cents2frac(double cents, int den);
int diatonic(int p);
int equaltemp(int n, double m, double scale[]);
int mod(int n, int m);
int nearestdiat(double pitch);

/* extern (sagfont.c) */
int output_sagittal_fmt(FILE *fp);

/* extern (sagittal.[ch]) */
//int sagcode2abc (char *dest, int code, char nominal, int precision, int pyth);
extern const int primes[NPRIMES]; /*  */

/* local */

int get_proc_command(char *s, FILE *fin, FILE *fout);
int prepare();
int normalizesagittal(char *s);
int bestfifthindex();
int clean_string(char *s);
int set_12out_et(int best /* fifth index */);
int set_pythagorean();
int set_meantone();
int includefile(FILE *fout,char name[]);
int preprocess(FILE *fin, FILE *fout, int outfor);
int parse_sagittalapprox(char *s);
int prepare_sagittal();
int set_temper();
int sagittalprocmac(char s[], int lim, int outfor);
int sagittalprocabc(char s[], int lim);
static char *fgetline(char s[],int lim,FILE *ap_arquivo);
double pitch2freq(int pitch);
double scl2cents(char *s);
int abc(char *s, int n, int typ);
int alias(char *s, int n, int i);
int espacos(char *s);
int fixabc(char *s);
int get_option(char *s);
int initialize();
int input(FILE *fp);
static int iseol(char c);
int microabc(char *s, int p);
int freq2abc(char *s, double f);
double abc2freq(char *s);
int output(FILE *fp);
int output_psacc(FILE *fd);
int output_scala();
int output_testfile();
int output_scalakbm();
int output_snt(FILE *fd); /* when fd==NULL, write to snt_file */
int output_timidity();
int print_ps_accid(FILE *fd);
int readscalafile(char *fname);
int remove_accid(char *s);
int remove_bad_updown(char *s);
int set_basefreq();
int storenumerator(int n);
int syncalias();
unsigned microval(int n, int d);

#endif
