/* sagittal.c --- Sagittal support for microabc */

/* This program is part of microabc (C) 2006-2008 Hudson Lacerda */

/* Copyright  (C)  2006-2008  Hudson F. M. Lacerda <hfml @ brfree.com.br> */

/* Author: Hudson F. M. Lacerda <hfml @ brfree.com.br> */
/* Maintainer: Hudson F. M. Lacerda <hfml @ brfree.com.br> */
/* URL: http://br.geocities.com/hfmlacerda/ */

/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 2 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software */
/* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "sagittal.h"
#include "sagdata.h"

extern sagstruct sagittal[NSAGCHARS]; /* in sagdata.c */
/* int isspace(int c); // from ctype.h */

const int primes[NPRIMES] = { /* On change, update NPRIMES in sagittal.h */
     2,   3,   5,   7,
    11,  13,  17,  19,
    23,  29,  31
};

int sagoctave[2] = {SAG_DEFAULT_OCT,0}; /* default octave index */

/*

TODO:
-- interface in microabc to substitute sagittal code acidentals
-- interface in microabc to substitute sagittal macros
-- treat sagittal accidental/nominal/octave

*/

/*
#ifdef TESTING
  int main()
  {

  int n, code, c;
  char s[BUFSIZ], t[BUFSIZ];
  char *p;
  while ((fscanf(stdin,"%s",s)>0)) {
  if(sscanf(s,"^/%d%c",&code,&c)==2) {
  sagcode2abc(t,code,c,4096);
  printf("%s",t);
  }
  if(sscanf(s,"_/%d%c",&code,&c)==2) {
  sagcode2abc(t,code,c,4096);
  printf("%s",t);
  }
  printf("\n");
  }
  }
#endif
*/

int sagcode2abc (char *dest, int code, char nominal,
		 int precision, int pyth, double temper[])
{

    /* try convert abcsagcode to abc microtone -- return 1 on success,
       0 on failure */

    double num;
    int den;
    int up, dn;
    char accid[64];
    double x;
    double f1, f2, f3, f4;

    accid[0]=accid[1]=accid[2]='\0';

    /* don't convert accidental off range */
    if(code<MINSAGCODE || code>MAXSAGCODE) {
	*dest='\0';
	return 0;
    }
    /* don't accept bad nominal */
    if( (tolower(nominal)<'a' || tolower(nominal)>'g') ) {
	*dest='\0';
	return 0;
    }

    /* convert data (from pitchbend in cents to semitone fraction) */
    
    if(temper[0]==0.0 && temper[1]==0.0) /*assume pure; TODO: improve*/
	num = sagittal[code].cents * precision / 100.0;
    else {
	num = sagtemper(code,temper) * precision / 100.0;
    } 
	
    den = precision;

    if(pyth) /* adjust nominals */ {
	adjustnominals (&num,den,nominal,temper);
    }

    /* TODO: handle Sagittal accents */
    /*
     */

    /* treat small pitchbends */
    if(round(num)==0.0) {
	sprintf(dest,"=%c",nominal);
	return 1;
    }
    else if(num<0 && num>=(-den)) {
	sprintf (dest, "_%d/%d%c",(int)round(fabs(num)),den,nominal);
	return 1;
    }
    else if(num>0 && num<=den) {
	sprintf (dest, "^%d/%d%c",(int)round(num),den,nominal);
	return 1;
    }

    if(round(num)==0.0)
	strcpy(accid,"=");
    else if(num>0)
	strcpy(accid,"^");
    else 
	strcpy(accid,"_");

    up=dn=0;


    /* adjust large pitchbends with double accidental */
    /* TODO: Check that `%%MIDI snt' IS NOT used */
    
    /*
    if(pyth) {
	if(num>(2*den)) {
	    strcpy(accid,"^^");
	    num-=den;
	}
	else if((-num)>(2*den)) {
	    strcpy(accid,"__");
	    num+=den;
	}
    }
    */

    /* TODO: improve code below.
       Bad values in temper[] can generate
       infinite loops */
    ///
    //    fprintf(stderr,"temper[1]+FIFDIF=%lf\n",temper[1]+FIFDIF);
    ////

    f1= (5 *(temper[1]+FIFDIF) * den/100.0);
    f2= (2 *(temper[1]+FIFDIF) * den/100.0);
    f3= (-5 *(temper[1]+FIFDIF) * den/100.0);
    f4= (-2 *(temper[1]+FIFDIF) * den/100.0);


    x=num;

    if(pyth) {
	/* TODO: FIND ANOTHER WAY TO TEMPER TO CAN DEAL WITH, FOR INSTANCE, 5-EDO */
	/* treat very large pitchbends when using pythagorean nominals */
	while(round(num)>(den*2)) { /* more than 200 cents up */
	    if(num>2*x) {
		fprintf(stderr,"\nERROR (fixme): "
			"sagcode2abc() -- can't adjust pitchbend!!! (1)\n");
		break;
	    }
	    switch (nominal) {
	    case 'E':
		nominal='F';
		num-=(den + f1);
		break;
	    case 'e':
		nominal='f';
		num-=den + f1;
		break;
	    case 'B':
		nominal='c';
		num-=den + f1;
		break;
	    case 'b':
		nominal='c';
		up=1;
		num-=den + f1;
		break;
	    case 'G':
		nominal='A';
		num-=(2*den) + f2;
		break;
	    case 'g':
		nominal='a';
		num-=(2*den) + f2;
		break;
	    default:
		nominal+=1;
		num-=((2*den) + f2);
	    }
	}

	x=num;

	while(round(num)<(-den*2)) { /* more than 200 cents down */
	    if(num<2*x) {
		fprintf(stderr,"\nERROR (fixme): "
			"sagcode2abc() -- can't adjust pitchbend!!! (2)\n");
		break;
	    }
	    switch (nominal) {
	    case 'F':
		nominal='E';
		num+=den + f3;
		break;
	    case 'f':
		nominal='e';
		num+=den + f3;
		break;
	    case 'C':
		nominal='B';
		num+=den + f3;
		dn=1;
		break;
	    case 'c':
		nominal='B';
		num+=den + f3;
		break;
	    case 'A':
		nominal='G';
		num+=(2*den) + f4;
		break;
	    case 'a':
		nominal='g';
		num+=(2*den) + f4;
		break;
	    default:
		nominal-=1;
		num+=(2*den) + f4;
	    }
	}
    } /* endif pyth */


    /* output adjusted pitchbend/nominal/up/dn */
    if(round(num)==0) {
	sprintf (dest,"%s%c",accid,nominal);
    }
    else if(num<0) {
 	sprintf (dest, "%s%d/%d%c",accid,(int)round(fabs(num)),den,nominal);
    }
    else if(num>0) {
	sprintf (dest, "%s%d/%d%c",accid,(int)round(num),den,nominal);
    }

    if(up && !dn)
	strcat (dest,"'"); /* ABC octave up */
    else if(dn && !up)
	strcat (dest,","); /* ABC octave down */

    return 1;
}

int adjustpythagoras (double *num, int den, char nominal)
{
    int c;
    double n;
    char *x;

    const char pylow[8] = "fcgdaeb";
    const char pyupp[8] = "FCGDAEB";

    /* index of middle reference pitch (F=0, C=1, G=2, D=3, A=4 etc.) */
    const int m = REFNOMINAL;


    x=NULL;
    c=0;
    
    x = strchr(pylow,nominal);
    if(x!=NULL)
	c=(x-pylow)-m;

    x = strchr(pyupp,nominal);
    if(x!=NULL) 
	c=(x-pyupp)-m;

    n = *num + (FIFDIF * c * (double)den/100.0);
    *num = n;

    return 1;
}

/* adjust Sagittal pitch into arbitrary temperament (using a reference) */
int adjustnominals (double *num, int den, char nominal, double temper[])
{
    int c;
    double n;
    char *x;
    const int m = REFNOMINAL; /* index of middle reference pitch */
    const char pylow[8] = "fcgdaeb";
    const char pyupp[8] = "FCGDAEB";

    x=NULL;
    c=0;
    
    x = strchr(pylow,nominal);
    if(x!=NULL)
	c=(x-pylow)-m;

    x = strchr(pyupp,nominal);
    if(x!=NULL) 
	c=(x-pyupp)-m;

    /* TODO: check if it is needed to take in account the octave size */
    n = *num + ((temper[1]+FIFDIF) * c * (double)den/100.0);
    *num = n;

    return 1;
}

double sagtemper(int code,double temper[])
{
    double d,cents;
    int i;
    cents = sagittal[code].cents;

    for(i=0; i<NPRIMES; i++) {
	d = primefactor(sagittal[code].num, sagittal[code].den, primes[i]);
	cents += d*temper[i];
    }
    return cents;
}

int setsagpitch(sagpitch *pitch,
		const char name[],
		const char nominals[],
		const unsigned char asciiformat)
{
    char *p;
    char s[MAXSAGNAME], t[MAXSAGNAME], u[8];
    int i, found, c, octaveread, oct;
    sagpitch sp;

    cleansagpitch(&sp);

    /* copy name */
    strncpy(sp.name, name, MAXSAGNAME-1);
    strncpy(s, name, MAXSAGNAME-1);

    /* find nominal -- it can be any uppercase letter, or, if there
       are not any uppercase letters, it can be a number */
    /* TODO: HANDLE XMNUWY AND |||| !!!! ACCIDENTALS */


    /* Look for the indicator of DONTSHOW accidental in PS score */
    p = s;
    while((*p)==DONTSHOW || isspace(*p)) {
	/* get left accents */
	if((*p)==DONTSHOW) { /* do not show accidental in score */
	    sp.accidental.visible = 0;
	    *p = ' '; /* remove DONTSHOW indicator */
	    break;
	}
	p++;
    };


    /* TODO: Handle octaves as ,, or `` here */
    /* Octaves as numbers are handled more below */
    octaveread=0;
    oct=0;
    i=strlen(p);
    while(--i>=0) {
	if(p[i]==SAGOCTAVEUP)
	    oct++;
	else if(p[i]==SAGOCTAVEDN)
	    oct--;
	else if(isspace(p[i]))
	    continue;
	else
	    break;
	octaveread=1;
	p[i]=' ';
    }

    /* Actually, just look for a char in nominals[] */
    i=0;
    p=s;
    found = 0;
    while((nominals[i])!='\0') {
	p=strchr(s,nominals[i]) ;
	if(p!=NULL) {
	    ////printf("nominals[i]={%c},*p={%c} ",nominals[i],*p);
	    sp.nominal = *p;
	    *p = ' '; /* remove nominal */
	    found=1;
	    break;
	}
	i++;
    }

    /*else if (  )
      nominal is a number (possibly more than one digit?) */
    if(found==0)
	return 0; /* name is not a valid sagittal pitch */

    /* find octave instance -- it must be a number, possibly preceded
       by ':' or '.' or a blank space */
    p++; /* now it points to after the (one-char) nominal */
    while(isdigit(*p)==0 && (*p)!='\0') /* skip non-digits */
	p++;
    /* now p points to either '\0' or a digit */

    if((*p)=='\0') /* no octave index */
	//// TODO: CHECK!
	i = sagoctave[0]; /* default octave index */
    else { /* found octave index */
	if(octaveread) /* cannot have octave as both number and `` ,, */
	    return 0;
	sscanf(p,"%d",&i); /* get number */

	/* if sagoctave[0] is zero, treat '-' as negative and '.' as
	   number separator -- this is the old behaviour */
	//// TODO: IMPROVE THIS CONTROL !!!
	if (sagoctave[0] == 0) {
	    if(  (*(p-1))=='-') {
		/*TODO: check index of pointer*/
		if (( *(p-2)==':' || *(p-2)=='.' || isspace(*(p-2)) )
		    ) {
		    p--;
		}
		i=(-i); /* negative octave index */
		p--;
	    }
	    if((*(p-1))==':' || (*(p-1))=='.' || isspace(*(p-1))) {
		/* is a separator */
		p--;
	    }
	}
	else { /* octaves cannot be negative, and '.' is always
		  sagittalgrave accent */
	    if((*(p-1))==':' || isspace(*(p-1))) {
		/* is a separator */
		p--;
	    }
	}
	*p='\0'; /* close s[] just before octave */
    }

    /* store octave index */
    if(octaveread) /* `` or ,, */
	sp.octave = oct + sagoctave[1];
    else
	sp.octave = i - sagoctave[0] + sagoctave[1];

    /* find sagittal accidental */
    /* find sagittal accents ( ' or . ) */

    p=s;
    i = 0;
    while((i<MAXSAGACCENTS) && (((*p)=='\'' || (*p)=='.') || isspace(*p))) {
	/* get left accents */
	if(isspace(*p)==0)
	    sp.accidental.left[i++] = *p;
	p++;
    };
    strcpy(s,p); /* skip left accents */

    if ((p=strchr(s,'\''))!=NULL)
	; /* p points to the accent */
    else if((p=strchr(s,'.'))!=NULL)
	; /* p points to the accent */
    else
	p=s; /* p points to main accidental (possibly with right accents) */

    i=0;
    while(i<MAXSAGACCENTS && ((*p)=='\'' || (*p)=='.')) {
	/* get right accents */
	sp.accidental.right[i] = *p;
	*p = '\0'; /* close string after main accidental */
	p++;
	i++;
    };

    /* check if remainder is a valid sagittal accidental */
    t[0]=u[0]=0;
    if(strlen(s))
	i = sscanf(s,"%s %7s",t,u);
    else
	i = 1, t[0]=0, u[0]=0;
    if(i==0 || i>1) /* cannot have spaces inside or be void */
    	return 0;
    /* the main accidental is now stored in t[] */

    ///    fprintf(stderr," i= %d :%s:%s:%s:\n",i,s,t,u);
    
    /* with accidental?  */
    u[0]='\0';
    if(strlen(t) && sscanf(t,"%s",u) > 0) {
	/* lookup in sagdata to get decimal code */
	c=0;
	/* Decimal code before nominal */
	if(sscanf(t,"%d",&c) && c>=MINSAGCODE && c<=MAXSAGCODE)
	    i = c;
	/* ASCII name */
	else if((i = sagasciiname2code(t,asciiformat))==0) {
	    ///fprintf(stderr,"\nUNKNOWN SAGITTAL ACCIDENTAL: %s\n\n",t); /////
	    return 0; /* TODO: CHECK */
	}
    }
    else { /* no accidental, only a nominal */
	sprintf (sp.name , "%c",sp.nominal );
	i= SAGSPACE; ///// TODO: check this
    }
    strcpy(sp.accidental.name,t); ////
    sp.accidental.code = i;

    /* OK; copy to output struct */
    copysagpitch(pitch, &sp);

    return 1;
}

int copysagpitch(sagpitch *dest, sagpitch *orig)
{
    strcpy (dest->name, orig->name);
    dest->nominal = orig->nominal;
    dest->octave = orig->octave;
    copysagaccid(&(dest->accidental), &(orig->accidental));
    return 1;
}

int copysagaccid(sagaccid *dest, sagaccid *orig)
{
    int i;
    dest->code = orig->code;
    dest->visible = orig->visible;
    strcpy(dest->name,orig->name); //////
    for(i=0; i<MAXSAGACCENTS; i++) {
	dest->left[i] = orig->left[i];
	dest->right[i] = orig->right[i];
    }
    return 1;
}

int cleansagpitch(sagpitch *sag)
{
    int i;
    sag->accidental.code = 0;
    for(i=0; i<MAXSAGACCENTS; i++) {
	sag->accidental.left[i] = '\0';
	sag->accidental.right[i] = '\0';
    }
    strcpy (sag->name, "");
    strcpy (sag->accidental.name, ""); /////
    sag->nominal = '\0';
    sag->octave = 0;
    sag->accidental.visible = 1;
    return 1;
}

// TODO: put prototype in sagittal.h
int setsagaccid(sagaccid *acc, const char name[])
{
    // this function will interpret a sagittal accidental name
    // and store the accidental in the struct
    return 1;
}

int sagasciiname2code(const char *s, const unsigned char f)
{
    /* recognize core accidental as ascii name and return the code */
    int i;
    char a[MAXSAGNAME], b[MAXSAGNAME], *p;

    strcpy(b,s);
    i=strlen(b);

    a[0]='\0';

    if((p=strstr(b,"#"))!=NULL) {
	*p=' ';
	sscanf(b,"%s",a);
	strcat(a,"#");
    }
    else if((p=strstr(b,"x"))!=NULL) {
	*p=' ';
	sscanf(b,"%s",a);
	strcat(a,"x");
    }
    else if((p=strstr(b,"bb"))!=NULL) {
	*p= *(p+1)=' ';
	sscanf(b,"%s",a);
	strcat(a,"bb");
    }
    else if((p=strstr(b,"b"))!=NULL) {
	*p=' ';
	sscanf(b,"%s",a);
	strcat(a,"b");
    }
    else
	strcpy(a,b);

    for(i=MINSAGCODE; i<MAXSAGCODE; i++) {
	if(f & SHORTPURE)
	    if(strcmp(a,sagittal[i].shortpure)==0)
		return sagittal[i].code;
	if(f & LONGPURE)
	    if(strcmp(a,sagittal[i].longpure)==0)
		return sagittal[i].code;
	if(f & SHORTMIXED)
	    if(strcmp(a,sagittal[i].shortmixed)==0)
		return sagittal[i].code;
	if(f & LONGMIXED)
	    if(strcmp(a,sagittal[i].longmixed)==0)
		return sagittal[i].code;
	if(f & ALTERNATIVE)
	    if(strcmp(a,sagittal[i].alternative)==0)
		return sagittal[i].code;
    }
    return 0;
}

double primefactor(double num, double den, double base)
{
    double powe, div;

    powe=0.0;
    div=num;
    while( div==floor(div) ) {
	div=div/base;
	powe++;
    }
    div=den;
    while( div==floor(div) ) {
	div=div/base;
	powe--;
    }
    return powe;
}








