/* abc2macro.c --- a program to convert ABC code into a ABP file for microabc */

/* Copyright (C) 2006-2008 Hudson F. M. Lacerda <hfml @ brfree.com.br> */

/* Author: Hudson F. M. Lacerda <hfml @ brfree.com.br> */
/* Maintainer: Hudson F. M. Lacerda <hfml @ brfree.com.br> */
/* URL: http://br.geocities.com/hfmlacerda/ */

/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 2 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software */
/* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

//
// TODO:
// -- treat ABC-2.0 continuation lines
// -- treat key signatures and voices
// cannot deal with invisible bars like [|]
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define DFLAT   0
#define FLAT    1
#define NATURAL 2
#define SHARP   3
#define DSHARP  4
#define NONE    5

#define CHORDMARK '\a' /* a character unlikely to be found in a text file */
/* #define CHORDMARK '@' */

typedef struct {
    char nominal;
    int accidental;
    int num; /* for microtones */
    int den;
    int octave;
    char len[32];
} abcpitch;


/* typedef struct { */
/*     int num; /\* voice number *\/ */
/*     char name[BUFSIZ]; /\* voice name *\/ */
/*     abcpitch key[7]; /\* key signature for this voice *\/ */
/*     abcpitch accidstate[32]; /\* memorize accidentals inside a measure *\/ */
/* } abcvoiceinfo; */

static char *fgetline(char s[],int lim,FILE *ap_arquivo);
static int iseol(char c);
int espacos(char *s);
int abc2macro(FILE *fin, FILE *fout);
int remove_comments(char *s, char *com);
int abc2sag(char s[], abcpitch *pitch);
int abc2abc(char s[], abcpitch *p);
int abc2str(char s[], abcpitch *p);
int cpyabcpitch(abcpitch *d, abcpitch *o);
int clearabcpitch(abcpitch *d);
int replacestr(char *s, const char *m, const char *r);
int isbar(char c);
int remove_bad_updown(char *s);
int fixabc(char *s);
int remove_linecontin(char *s);

int to_sagittal;
int split_chords;

int main(int argc, char **argv)
{
    if(argc>1) {
	split_chords=1;
	to_sagittal=0;
    }
    else {
	split_chords=0;
	to_sagittal=1;
    }
    abc2macro(stdin,stdout);
    return 0;
}

int abc2macro(FILE *fin, FILE *fout)
{
    int i,d,n,c,len,m;
    int in_tuneheader=0;
    int in_tune=0;
    int in_tunebody=0;
    int in_chord=0;
    int end_of_measure=0;
    //int in_inlinefield=0;
    //int in_annotation=0;
    int in_decoration=0; /* warning: ++ and !! formats */
    //int in_comment=0; /* comment marks can be escaped in strings! */
    char s[BUFSIZ], t[BUFSIZ], u[BUFSIZ], line[BUFSIZ];
    char com[BUFSIZ], note[BUFSIZ], v[BUFSIZ];
    char basestr[BUFSIZ];
    char *p, *q;
    const char chmark[4]={'[',CHORDMARK,']',0};
    abcpitch pitch;
    int nchords, nchordnotes;
    struct {
	int n;
	abcpitch pitch[16];
    } chords[16];

    basestr[0]='\0';
    nchords=nchordnotes=0;

    clearabcpitch(&pitch);
    for(n=0;n<16;n++)
	for(m=0;m<16;m++)
	    clearabcpitch(&chords[n].pitch[m]);

    while((fgetline(line,BUFSIZ,fin)!=NULL)) { /* FOR EACH INPUT LINE */
	/* discard EOL characters (for cross-plataform compatibility) */
	p=strchr(line,10); if(p!=0) *p = '\0'; /* descarta eol */
	p=strchr(line,13); if(p!=0) *p = '\0'; /* descarta eol */

	strcpy(s,line);
	remove_comments(s,com);
	remove_linecontin(s);
	////fprintf(stderr,"Was:\n%s\nIs now:\n%s\n",line,s);
	if(in_tunebody) {
	    p=s;
	    len=strlen(s);
	    while((c = *p)!='\0') {
		if(c==']'  && in_chord && split_chords) {
		    ///in_chord=0; // not here; when storing/outputing chars
		    chords[nchords-1].n=nchordnotes;
		    nchordnotes=0;
		    //chords[nchords].n = 0; /* zero notes in _next_ chord */
		}

		/* try catch alone-in-one-line fields -- cannot have spaces at left!!! */
		if(*(p+1)==':') {
		    if(isupper(c) || c=='w' || c=='d' || c=='s' || c=='r') {
			fprintf(fout,"%s",s);
			break;
		    }
		}

		if(c=='|' || c=='&'
		   || (c== ':' &&!isdigit(*(p+1)) )
		   || end_of_measure ) {

		    if(split_chords==1) {
			end_of_measure=1;			

			///fprintf(stderr,"\t;%s; (%d chords)\n",basestr,nchords);

			nchordnotes=0;
			for(m=0; m<nchords; m++)
			    nchordnotes =  (nchordnotes > chords[m].n) ? nchordnotes : chords[m].n;
			v[0]='\0';
			for(m=0;m<nchordnotes-1;m++) {
			    strcpy(v,basestr);
			    ///fprintf(stderr,"{%s}",v);
			    for(n=0;n<nchords;n++) {
				if(chords[n].n < m) {
				    v[0]='\0';
				    break;
				}
				note[0]='\0';
				i=abc2abc(note,&chords[n].pitch[m]);
				clearabcpitch(&chords[n].pitch[m]);
				if(n==0)
				    fprintf(fout,"&");
				if(i) {
				    replacestr(v,chmark,note);
				    ////fprintf(stderr,"{(1)%s}",v); ////
				}
				else {
				    replacestr(v,chmark,"x");
				    ///fprintf(stderr,"{(2)%s}",v); /////
				}

			    }
			    fprintf(fout,"%s",v);
			}

			basestr[0]=v[0]=u[0]='\0';
			s[0]='\0'; /* <<=== TODO: check this */
			nchords=0;
			nchordnotes=0;
			in_chord=0;
			end_of_measure=0;

			for(n=0;n<16;n++)
			    for(m=0;m<16;m++)
				clearabcpitch(&chords[n].pitch[m]);

			while(isbar(c) || isdigit(c)) {
			    p++;
			    fputc(c,fout);
			    c=*p;
			}
			/*
			  if(c=='\0')
			  //c=*--p;
			  */
		    }
		    u[0]=v[0]=basestr[0]='\0';
		}

		/* try get inline field */
		if(c=='[') {
		    if((p+2<s+len) /* <== TODO: Check!!!*/
		       && *(p+2)==':') {
			fputc(c,fout);
			p++;
			while((c = (*p)) != '\0' ) {
			    fputc(c,fout);
			    if(c==']') {
				c = *(++p);
				break;
			    }
			    c=*++p;
			}
			
			/* TODO: parse [V:] and [K:] */
			
			/// fprintf(fout,"[r:inline field]");
			///    continue;
		    }
		    else { /* is chord? */
			in_chord=1;
			nchords++;
			nchordnotes=0; //////////////////////
			///c=*++p;
		    }
		}

		/* try get special barline |] or [| ??? */
		;

		/* try get annotation/gchord */
		if(c=='"') { /* TODO: treat escaped \" */
		    fputc(c,fout);
		    p++;
		    while((c = (*p)) != '\0' ) {
			fputc(c,fout);
			if(c=='"') {
			    c = *(++p);
			    break;
			}
			p++;
		    }
		    /// fprintf(fout,"[r:gchord]");
		    
		    /* TODO: What about multiline annotations?? */
		    
		}

		/* try get try get decoration. !+! and +plus+ are both
		   valid! */
		if(c=='!' || c=='+') {
		    q=p;
		    in_decoration = c; /* N.B. */
		    if(c=='!') {
			/*
			  When encountering a !, scan forward. If you find
			  another ! before encountering any of "|[:]", a
			  space, or the end of line, then you have a
			  decoration, otherwise it is a line break.
			*/
			while(q++<s+len) {
			    if(isspace(*q)
			       || *q=='|'
			       || *q=='['
			       || *q==':'
			       || *q==']'
			       || *q=='\0') {
				in_decoration=0;
				break;
			    }
			    else if(*q == '!')
				break;
			}
		    }
		    else
			in_decoration='+';
		    if(in_decoration) {
			fputc(c,fout);
			p++;
			while((c = (*p)) != '\0' ) {
			    if(c==in_decoration) {
				fputc(c,fout);
				p++;
				c = *p;
				break;
			    }
			    fputc(c,fout);
			    p++;
			}
			///	fprintf(fout,"[r:deco]");
			/* TODO: What about multiline annotations?? */
		    }
		    in_decoration=0;
		}

		clearabcpitch(&pitch);
		pitch.accidental=NONE;
		if(c=='=' || c=='^' || c=='_') { /* abc accidental */
		    if(c=='=')
			pitch.accidental=NATURAL;
		    else if (c=='^') {
			if(*(p+1)=='^') {
			    pitch.accidental=DSHARP;
			    p++;
			}
			else
			    pitch.accidental=SHARP;
		    }
		    else if (c=='_') {
			if(*(p+1)=='_') {
			    pitch.accidental=DFLAT;
			    p++;
			}
			else
			    pitch.accidental=FLAT;
		    }
		    p++;

		    /* treat microtone accidentals */
		    pitch.num=1;
		    pitch.den=1;
		    if(*(p)=='/' || isdigit(*(p))) {
			if(*p=='/') {
			    if((n=sscanf(p,"/%d",&pitch.den))==0)
				pitch.den=2;
			}
			else if((n=sscanf(p,"%d/%d",&pitch.num,&pitch.den))) {
			    if(n==1)
				pitch.den=2;
			}
			else {
			    pitch.num=1;
			    pitch.den=1;
			    ///fprintf(fout,"[r:nomicrotone]");
			}
			while(isdigit(*p) || *p=='/')
			    p++;
		    }
		    c=*p;
		    /* ///  fprintf(fout,"[r:acc=%d,num=%d,den=%d *p=%c]",
		       pitch.accidental,
		       pitch.num,
		       pitch.den,c); */
		}
		////fprintf(fout,"[r:%c]",*p);
		if((tolower(c) >= 'a' && tolower(c) <= 'g')
		   || c=='x' || c=='z') { /* abc pitch or rest */
		    pitch.nominal=c;
		    pitch.octave=0;
		    p++;
		    /* try get octave */;
		    while((c=*(p))=='\'') {
			pitch.octave++;
			p++;
		    }
		    while((c=*(p))==',') {
			pitch.octave--;
			p++;
		    }
		    ///c=*p;
		    n=0;
		    if(in_chord && split_chords) /* only store
						    duration for
						    chords to split */
			while((c=*(p))=='/' || isdigit(c)) {
			    pitch.len[n++]=c;
			    p++;
			}
		    pitch.len[n]='\0';

		    /* HERE THE ABC PITCH IS PARSED */
		    /* CONVERT IT TO MICROABC MACRO */

		    note[0]='\0';
		    abc2str(note,&pitch);

		    /* in in_chord==0, copy a 'x' rest to basestr[] */
		    if(in_chord==0 && split_chords==1) {
			sprintf(u,"x");
			strcat(basestr,u);
		    }
		    else {
			if(pitch.nominal != 'x' && pitch.nominal !='z') {
			    if(split_chords) {
				if(nchordnotes) {
				    cpyabcpitch(&chords[nchords-1].pitch[nchordnotes-1],
						&pitch);
				}
				if(nchords>0 && nchordnotes==1) {
				    sprintf(u,"%c",CHORDMARK);
				    strcat(basestr,u);
				}
				nchordnotes++;
			    }
			}
			else {/* rest */
			    sprintf(u,"x");
			    strcat(basestr,u);
			}
		    }

		    if(nchordnotes<=1 || split_chords==0) {
			if(pitch.nominal != 'x' && pitch.nominal !='z') {
			    fprintf(fout,"%s",note);
			}
			else /* rest */
			    fprintf(fout,"%s",note);
		    }

		    clearabcpitch(&pitch);
		}
		else { /*  TODO: omit in-chord decorations */
		    c=*p;
		    if(c=='\0') {
			///c=*--p;
			break;
		    }
		    //fprintf(fout,"(*%d*)",c==0);
		    /* try deal with line continuation (omit it) */
		    /*if(split_chords && c=='\\' && espacos(p+1)) {
		      p++;
		      continue;
		      }*/

		    /* not a note or rest -- just copy */
		    if(split_chords==0) {
			in_chord=0;
		    }
		    if(c=='[' && in_chord)
			;
		    else if(c==']' && in_chord)
			in_chord=0;
		    else
			fputc(c,fout);
		    sprintf(u,"%c",c);
		    strcat(basestr,u);
		    p++;
		    if(*p=='\0') {
			c=*--p;
			break;	   
		    }
		}
	    }
	}
	else
	    fprintf(fout,"%s",s); /* output non-interpreted non-music line */

	if( (sscanf(s,"X:%d",&d)) ) {
	    in_tuneheader=1;
	    in_tune=1;
	}
	if( sscanf(s,"K:%s",t) ) {
	    /* TODO: Change key in all voices */

	    if( in_tuneheader ) {
		in_tuneheader=0;
		in_tunebody=1;
		in_tune=1;
	    }
	}
	if((espacos(line) && !in_tuneheader)) {
	    in_tune=0;
	    in_tuneheader=0;
	    in_tunebody=0;
	    nchords=0;
	    nchordnotes=0;
	    in_chord=0;
	    note[0]='\0';
	    basestr[0]=v[0]=u[0]='\0';
	    clearabcpitch(&pitch);
	}

	/* output comment and EOL */
	fprintf(fout,"%s\n",com);

    }
    return 1;
}






static char *fgetline(char s[],int lim,FILE *ap_arquivo)
{
    int c,i,c2;

    s[0]=c='\0';

    for(i=0; i<lim-1
	    && (c=fgetc(ap_arquivo))!=EOF
	    && (!iseol(c));i++)
	s[i]=c;
    if(iseol(c)) { // TODO: check buffer full ??
	s[i]=c;
	i++;
	if(c==13 && i<lim-1) { // treat EOL format for DOS
	    if((c2=fgetc(ap_arquivo))!=EOF && c2==10) {
		s[i++]=c2;
		c=c2;
	    }
	    else
		ungetc(c2,ap_arquivo);
	}
    }
    s[i]='\0';

    if(c==EOF && strlen(s)==0)
	return NULL;

    return s;
}

static int iseol(char c)
{
    switch(c) {
    case 10:
    case 13:
	return 1;
    default:
	return 0;
    }
}

int espacos(char *s)
{
    //// extern int isspace(int c);
    int i=0;
    while(s[i]!='\0')
	if(!isspace(s[i++]))
	    return 0;
    return 1;
}

int remove_comments(char *s, char *com)
{
    char *p=NULL;
    char *q=NULL;
    com[0]='\0';
    p=strchr(s,'%');
    while(p!=NULL) {
	if(p==s || ((p>s) && *(p-1)!='\\')) {
	    strcpy(com,p);
	    *p='\0';
	    return 1;
	}
	q=p+1;
	if(*q=='\0')
	    return 1;
	p=strchr(q,'%');
    }
    return 0;
}

int remove_linecontin(char *s)
{
    char *p=NULL;
    char *q=NULL;

    p=strchr(s,'\\');
    while(p!=NULL) {
	if( espacos(p+1) ) { /* assume comments removed */
	    *p='y';
	    return 1;
	}
	q=p+1;
	if(*q=='\0')
	    return 1;
	p=strchr(q,'\\');
    }
    return 0;
}

int cpyabcpitch(abcpitch *d, abcpitch *o)
{
    d->nominal = o->nominal;
    d->accidental = o->accidental;
    d->num =  o-> num; 
    d-> den =  o->  den;
    d->octave =  o->  octave;
    strcpy(d->len, o->len);
    return 1;
}

int clearabcpitch(abcpitch *d)
{
    d->nominal = ' ';
    d->accidental = NONE;
    d->num = 1; 
    d-> den = 1;
    d->octave = 0;
    d->len[0] = '\0';
    return 1;
}

int isbar(char c)
{
    switch(c) {
    case '|':
    case ']':
	//    case '[': /* TODO: treat [1 [2 bars */
    case ':':
	return 1;
    }
    return 0;
}

int replacestr(char *s, const char *m, const char *r)
{
    int n,mlen,rlen;
    char *p, *q;
    char t[BUFSIZ];
    t[0]='\0';
    n=0;
    mlen=strlen(m);
    rlen=strlen(r);

    p=strstr(s,m);
    if(p!=NULL) {
	q=s;
	while(q<p)
	    t[n++]=*q++;
	t[n]='\0';
	strcat(t,r);
	///p+=mlen;
	n+=rlen;
	q=p+mlen;
	while(*q!='\0')
	    t[n++]=*q++;
	*q='\0';
    }
    t[n]='\0';
    strcpy(s,t);
    return 1;
}

int abc2str(char s[], abcpitch *p)
{
    if(to_sagittal)
	abc2sag(s,p);
    else
	abc2abc(s,p);
    return 1;
}

int abc2abc(char s[], abcpitch *p)
{
    char t[BUFSIZ];
    char u[BUFSIZ];
    int c,o;

    s[0]='\0';
    t[0]='\0';
    u[0]='\0';

    c = p->nominal;

    if(tolower(c)<'a' || tolower(c)>'g')
	if(c!='x' && c!='z')
	    return 0;

    /* visible or invisible rest */
    if(c == 'z' || c == 'x') {
	sprintf(s,"%c",c);
	return 1;
    }

    /* TODO: treat microtones */
    switch (p->accidental) {
    case SHARP:
	strcpy(t,"^");
	break;
    case FLAT:
	strcpy(t,"_");
	break;
    case NATURAL:
	strcpy(t,"=");
	break;
    case DSHARP:
	strcpy(t,"^^");
	break;
    case DFLAT:
	strcpy(t,"__");
	break;
    case NONE:
	strcpy(t,"="); /* hidden? natural */
	break;	
    default:
	t[0]='\0';
    }

    if(p->num!=1 || p->den!=1) {
	sprintf(u,"%d/%d",p->num,p->den);
	strcat(t,u);
    }

    u[0]='\0';
    o = p->octave;
    while(o>0) {
	strcat(u,"'");
	o--;
    }
    while(o<0) {
	strcat(u,",");
	o++;
    }

    sprintf (s, "%s%c%s", t,c,u);

    fixabc(s);

    strcat(s,p->len);

    return 1;
}

int abc2sag(char s[], abcpitch *p)
{
    char t[BUFSIZ];
    int c,o,n,d;
    s[0]='\0';
    t[0]='\0';

    c = p->nominal;
    n = p->num;
    d = p->den;

    /* visible or invisible rest */
    if(c == 'z' || c == 'x') {
	sprintf(s,"%c",c);
	return 1;
    }

    switch (p->accidental) {
    case SHARP:
	if(n==1 && d==2)
	    strcpy(t,"^");
	else if(n==3 && d==2)
	    strcpy(t,"^#");
	else
	    strcpy(t,"#");
	break;
    case FLAT:
	if(n==1 && d==2)
	    strcpy(t,"v");
	else if(n==3 && d==2)
	    strcpy(t,"vb");
	else
	    strcpy(t,"b");
	break;
    case NATURAL:
	strcpy(t,"e");
	break;
    case DSHARP:
	strcpy(t,"x");
	break;
    case DFLAT:
	strcpy(t,"bb");
	break;
    case NONE:
	strcpy(t,":e"); /* hidden natural */
	break;	
    default:
	;
    }

    o = p->octave;
    if(islower(c)) {
	c = toupper(c);
	o++;
    }
    if(c!='x' && c!='z')
	sprintf (s, "[%s%c%d]", t,c,o+5);
    else
	sprintf (s, "%c",c);
    strcat(s,p->len);
    return 1;
}


int fixabc(char *s) /* fix bad ABC pitches like C' C', c,, C'' */
{
    int c,i,n,oct,flag;
    char *p;
    char up0[8], down0[8];
    char up[8]="'", down[8]=",";

    /* remove all blank spaces */
    while((p=strchr(s,' '))!=NULL) {
	while(*(p+1)!='\0') {
	    *p = *(p+1);
	    p++;
	}
	*p='\0';
    }

    /* fix octaves as numbers */
    n=strlen(s);

    /* find number after nominal */
    flag=0;
    for(i=0;i<n;i++) {
	c = s[i];
	if(tolower(c)>='a' && tolower(c)<='g')
	    flag=1;
	else if(flag) {
	    if(isdigit(c) || c=='-')
		break;
	}
    }
    oct=0;
    if(i<n) { /* there was a bad octave number -- get it! */
	sscanf(s+i,"%d",&oct);
    }
    s[i]='\0'; /* remove number */
    while(oct!=0) { /* put '' or ,, */
	if(oct>0) {
	    strcat(s,"'");
	    oct--;
	}
	else { /* oct<0 */
	    strcat(s,",");
	    oct++;
	}
    }

    /* temporarily change updown to ABC ' and , */
    strncpy(up0,up,7);
    strncpy(down0,down,7);
    strcpy(up,"'");
    strcpy(down,",");

    remove_bad_updown(s);

    /* reset updown back */
    strncpy(up,up0,7);
    strncpy(down,down0,7);

    for(c='A';c<='G';c++) {
	p=strchr(s,c);
	if(p==NULL)
	    continue;
	if(strchr(s,'\'')!=NULL) {
	    *p=tolower(c);
	    s[strlen(s)-1]='\0'; /* remove trailing ' */
	}
    }
    for(c='a';c<='g';c++) {
	p=strchr(s,c);
	if(p==NULL)
	    continue;
	if(strchr(s,',')!=NULL) {
	    *p=toupper(c);
	    s[strlen(s)-1]='\0'; /* remove trailing , */
	}
    }
    return 1;
}

int remove_bad_updown(char *s)
{
    char *c1, *c2;
    char *p, *u;
    int s1, s2;
    int i, pisup,x;
    char up[8]="'", down[8]=",";

    for(x=0;x<5;x++) { /* awful hack to fix more pairs of bad updowns */
	pisup=0;
	s1=strlen(s);
	s2=strlen(up)+strlen(down);
	
	/* TODO: fix bad octave numbers??? */
	
	/* TODO: FIXME! By now, only removes first pair of up/down */
	if((c1=strstr(s,up))!=NULL && (c2=strstr(s,down))!=NULL) {
	    if(c1<c2)
		p=c1, u=c2, pisup=1;
	    else
		p=c2, u=c1, pisup=0;
	    for(i=0;i<s2;i++)
		*(p+i)=*(p+i+1);
	    s[s1-s2]='\0';
	}
    }
    return 1;
}

    
